from policyengine_us.model_api import *


class sc_retirement_deduction_indv(Variable):
    value_type = float
    entity = Person
    label = "South Carolina retirement deduction for eligible individuals"
    unit = USD
    reference = (
        "https://www.scstatehouse.gov/code/t12c006.php",  # SECTION 12-6-1170(A)(1)
        "https://dor.sc.gov/forms-site/Forms/IITPacket_2021.pdf#page=17",
    )
    definition_period = YEAR
    defined_for = StateCode.SC

    def formula(person, period, parameters):
        p = parameters(period).gov.states.sc.tax.income.subtractions.retirement
        age = person("age", period)
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        # line 1
        max_deduction_allowed = p.cap.calc(age)
        # line 2
        military_retirement_pay = person("military_retirement_pay", period)
        # line 3
        retirement_deduction_available = max_(
            max_deduction_allowed - military_retirement_pay, 0
        )
        # line 4
        retirement_income = (
            person("taxable_pension_income", period) * head_or_spouse
        )
        # line 5
        return min_(retirement_deduction_available, retirement_income)



from policyengine_us.model_api import *


class dc_ccsp_qualified_activity_or_need_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for DC Child Care Subsidy Program (CCSP) due to qualified activity or need"
    definition_period = MONTH
    defined_for = StateCode.DC
    reference = "https://osse.dc.gov/sites/default/files/dc/sites/osse/publication/attachments/DC%20Child%20Care%20Subsidy%20Program%20Policy%20Manual.pdf#page=8"

    def formula(spm_unit, period, parameters):
        qualified_activity_eligible = spm_unit(
            "dc_ccsp_qualified_activity_eligible", period
        )
        qualified_need_eligible = spm_unit(
            "dc_ccsp_qualified_need_eligible", period
        )
        return qualified_activity_eligible | qualified_need_eligible



from policyengine_us.model_api import *


class hi_income_tax_before_non_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii income tax before non-refundable credits"
    defined_for = StateCode.HI
    unit = USD
    definition_period = YEAR
    # Hawaii Tax Rate Schedules
    reference = "https://tax.hawaii.gov/forms/d_18table-on/d_18table-on_p13/"

    def formula(tax_unit, period, parameters):
        filing_status = tax_unit("filing_status", period)
        taxable_income = tax_unit("hi_taxable_income", period)
        p = parameters(period).gov.states.hi.tax.income.rates

        statuses = filing_status.possible_values

        income_tax = select(
            [
                filing_status == statuses.SINGLE,
                filing_status == statuses.SEPARATE,
                filing_status == statuses.JOINT,
                filing_status == statuses.SURVIVING_SPOUSE,
                filing_status == statuses.HEAD_OF_HOUSEHOLD,
            ],
            [
                p.single.calc(taxable_income),
                p.separate.calc(taxable_income),
                p.joint.calc(taxable_income),
                p.surviving_spouse.calc(taxable_income),
                p.head_of_household.calc(taxable_income),
            ],
        )
        alternative_tax = tax_unit(
            "hi_alternative_tax_on_capital_gains", period
        )

        alternative_tax_eligible = tax_unit(
            "hi_alternative_tax_on_capital_gains_eligible", period
        )
        return where(
            alternative_tax_eligible,
            min_(income_tax, alternative_tax),
            income_tax,
        )



from policyengine_us.model_api import *


class ks_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Kansas standard deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.ksrevenue.gov/pdf/ip21.pdf"
        "https://www.ksrevenue.gov/pdf/ip22.pdf"
    )
    defined_for = StateCode.KS

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ks.tax.income.deductions.standard
        filing_status = tax_unit("filing_status", period)
        base_amt = p.base_amount[filing_status]
        aged_blind_count = tax_unit("aged_blind_count", period)
        extra_amt = aged_blind_count * p.extra_amount[filing_status]
        return base_amt + extra_amt



from policyengine_us.model_api import *


class ssi(Variable):
    value_type = float
    entity = Person
    label = "SSI"
    documentation = "Supplemental Security Income"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/42/1382"

    def formula(person, period, parameters):
        if parameters(period).gov.ssa.ssi.abolish_ssi:
            return 0

        uncapped = person("uncapped_ssi", period)

        # Apply floor: SSI cannot be negative
        benefit = max_(0, uncapped)

        # Apply cap when spousal deeming: cannot exceed individual FBR
        # per 20 CFR §416.1163
        # This cap is necessary when spouse's gross income is just above $483:
        # - Deeming applies (uses couple FBR)
        # - After exclusions, countable may be low
        # - Benefit could exceed individual FBR without this cap
        deeming_applies = person("is_ssi_spousal_deeming_applies", period)
        p = parameters(period).gov.ssa.ssi.amount
        individual_max = p.individual * MONTHS_IN_YEAR
        capped_benefit = min_(benefit, individual_max)

        return where(
            deeming_applies,
            capped_benefit,
            benefit,
        )



from policyengine_us.model_api import *


class co_charitable_contribution_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Colorado charitable contribution subtraction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.colorado.gov/sites/tax/files/documents/DR0104AD_2022.pdf#page=1",
        "https://tax.colorado.gov/sites/tax/files/documents/DR_104_Book_2022.pdf#page=12",
        "https://casetext.com/regulation/colorado-administrative-code/department-200-department-of-revenue/division-201-taxation-division/rule-1-ccr-201-2-income-tax/rule-39-22-1044m-charitable-contribution-subtraction-for-taxpayers-claiming-the-federal-standard-deduction",
        # C.R.S. 39-22-104(4)(m)(1)
    )
    defined_for = "co_charitable_contribution_subtraction_eligible"

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.co.tax.income.subtractions.charitable_contribution
        charitable_contributions = add(
            tax_unit,
            period,
            ["charitable_cash_donations", "charitable_non_cash_donations"],
        )
        return max_(charitable_contributions - p.adjustment, 0)



from policyengine_us.model_api import *


class amt_kiddie_tax_applies(Variable):
    value_type = bool
    entity = TaxUnit
    definition_period = YEAR
    label = "Alternative Minimum Tax kiddie tax applies"
    documentation = "Whether the kiddie tax applies to the tax unit"

    def formula(tax_unit, period, parameters):
        age_head = tax_unit("age_head", period)
        p = parameters(period).gov.irs.dependent.ineligible_age
        young_head = (age_head != 0) & (age_head < p.non_student)
        no_or_young_spouse = tax_unit("age_spouse", period) < p.non_student
        return young_head & no_or_young_spouse



from policyengine_us.model_api import *


class vt_capital_gains_exclusion(Variable):
    value_type = float
    entity = TaxUnit
    label = "Vermont capital gains exclusion"
    unit = USD
    documentation = "Vermont excludes a portion of capital gains, calculated either as a flat amount or as a fraction of adjusted net capital gains, and limited by a fraction of federal taxable income."
    definition_period = YEAR
    defined_for = StateCode.VT
    reference = (
        "https://tax.vermont.gov/sites/tax/files/documents/IN-153-2022.pdf#page=1"  # 2022 Schedule IN-153 Vermont Capital Gains Exclusion Calculation
        "https://legislature.vermont.gov/statutes/section/32/151/05811"  # Titl. 32 V.S.A. § 5811(21)(B)(ii)
        "https://tax.vermont.gov/sites/tax/files/documents/IN-153%20Instr-2022.pdf"
    )

    def formula(tax_unit, period, parameters):
        # Get adjusted net capital gains, which is capped at 0
        adjusted_net_capital_gain = tax_unit(
            "adjusted_net_capital_gain", period
        )
        qualified_dividend_income = add(
            tax_unit, period, ["qualified_dividend_income"]
        )
        reduced_adjusted_net_capital_gain = max_(
            adjusted_net_capital_gain - qualified_dividend_income, 0
        )
        p = parameters(
            period
        ).gov.states.vt.tax.income.agi.exclusions.capital_gain
        # The flat exclusion is the less of a capped amount
        # or the actual amount of net adjusted capital gains
        flat_exclusion = min_(reduced_adjusted_net_capital_gain, p.flat.cap)
        # Get percentage exclusion
        percentage_exclusion = tax_unit(
            "vt_percentage_capital_gains_exclusion", period
        )
        # Filer can choose from flat or percentage exclusion.
        # Assume the filer will always choose the larger one
        chosen_exclusion = max_(flat_exclusion, percentage_exclusion)
        # The chosen exclusion should not exceed 40% of federal taxable income
        federal_taxable_income = tax_unit("taxable_income", period)
        cap = federal_taxable_income * p.income_share_cap
        return min_(chosen_exclusion, cap)



from policyengine_us.model_api import *


class hi_regular_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Hawaii regular exemptions"
    unit = USD
    documentation = (
        "https://files.hawaii.gov/tax/forms/2022/n11ins.pdf#page=20"
    )
    definition_period = YEAR
    defined_for = StateCode.HI

    def formula(tax_unit, period, parameters):
        exemptions_count = tax_unit("exemptions_count", period)
        p = parameters(period).gov.states.hi.tax.income.exemptions
        # Aged heads and spouses get an extra base exemption.
        person = tax_unit.members
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        aged = person("age", period) >= p.aged_threshold
        aged_head_spouse_count = tax_unit.sum(aged & head_or_spouse)
        total_exemption_count_including_aged = (
            exemptions_count + aged_head_spouse_count
        )
        return total_exemption_count_including_aged * p.base



from policyengine_us.model_api import *


class mn_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Minnesota taxable income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.state.mn.us/sites/default/files/2021-12/m1_21_0.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1_inst_21.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2022-12/m1_22.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-03/m1_inst_22.pdf"
    )
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        us_agi = tax_unit("adjusted_gross_income", period)
        adds = tax_unit("mn_additions", period)
        deductions = tax_unit("mn_deductions", period)
        exemptions = tax_unit("mn_exemptions", period)
        subs = tax_unit("mn_subtractions", period)
        return max_(0, us_agi + adds - deductions - exemptions - subs)



from policyengine_us.model_api import *


class lifeline(Variable):
    value_type = float
    entity = SPMUnit
    label = "Lifeline"
    documentation = "Amount of Lifeline phone and broadband benefit"
    definition_period = YEAR
    unit = USD
    defined_for = "is_lifeline_eligible"
    reference = "https://www.law.cornell.edu/cfr/text/47/54.403"

    def formula(spm_unit, period, parameters):
        # NB: Only one Lifeline benefit is available per SPM unit, per:
        # https://www.law.cornell.edu/cfr/text/47/54.409#c
        p = parameters(period).gov
        base_amount = p.fcc.lifeline.amount.standard
        household = spm_unit.household
        is_rural_tribal = and_(
            household, period, ["is_rural", "is_on_tribal_land"]
        )
        state_code = household("state_code_str", period)

        # Check for state-specific Lifeline programs
        STATES_WITH_LIFELINE = ["ca", "or"]
        for state in STATES_WITH_LIFELINE:
            state_params = p.states[state].fcc.lifeline
            if hasattr(state_params, "in_effect") and state_params.in_effect:
                base_amount = where(
                    state_code == state.upper(),
                    state_params.max_amount,
                    base_amount,
                )

        federal_monthly_amount = (
            base_amount
            + is_rural_tribal * p.fcc.lifeline.amount.rural_tribal_supplement
        )
        # Get TX supplement for the specific month (now properly defined as MONTH period)
        tx_supplement = spm_unit("tx_lifeline_supplement", period.first_month)
        total_monthly_amount = federal_monthly_amount + tx_supplement
        max_amount = total_monthly_amount * MONTHS_IN_YEAR
        phone_broadband_cost = add(
            spm_unit, period, ["phone_cost", "broadband_cost"]
        )
        return min_(phone_broadband_cost, max_amount)



from policyengine_us.model_api import *


class ca_ala_general_assistance_eligible_person(Variable):
    value_type = bool
    entity = Person
    label = "Eligible person for Alameda County General Assistance"
    definition_period = MONTH
    defined_for = "in_ala"
    reference = "https://www.alamedacountysocialservices.org/acssa-assets/PDF/GA-Policies/GA-Regulations.pdf#page=19"

    def formula(person, period, parameters):
        age_eligible = person("ca_ala_general_assistance_age_eligible", period)
        personal_property_eligible = person.spm_unit(
            "ca_ala_general_assistance_personal_property_eligible", period
        )
        immigration_status_eligible = person(
            "ca_ala_general_assistance_immigration_status_eligible", period
        )
        receives_ssi = person("ssi", period) > 0
        is_head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        return (
            age_eligible
            & personal_property_eligible
            & immigration_status_eligible
            & ~receives_ssi
            & is_head_or_spouse
        )



from policyengine_us.model_api import *


class ca_riv_share_payment(Variable):
    value_type = float
    entity = SPMUnit
    label = "Riverside County Sharing Households Assist Riverside's Energy program (SHARE) payment"
    unit = USD
    definition_period = MONTH
    defined_for = "ca_riv_share_eligible"
    reference = "https://riversideca.gov/utilities/residents/assistance-programs/share-english"

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.local.ca.riv.cap.share.payment
        electricity_expense = spm_unit(
            "pre_subsidy_electricity_expense", period
        )
        capped_electricity_payment = min_(electricity_expense, p.electricity)
        electricity_emergency_payment = spm_unit(
            "ca_riv_share_electricity_emergency_payment", period
        )

        water_expense = spm_unit("water_expense", period)
        capped_water_payment = min_(water_expense, p.water)

        trash_expense = spm_unit("trash_expense", period)
        capped_trash_payment = min_(trash_expense, p.trash)

        return (
            capped_electricity_payment
            + electricity_emergency_payment
            + capped_water_payment
            + capped_trash_payment
        )



from policyengine_us.model_api import *


class ia_taxable_income_consolidated(Variable):
    value_type = float
    entity = TaxUnit
    label = "Iowa taxable income for years on or after 2023"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://revenue.iowa.gov/media/2746/download?inline",
        "https://www.legis.iowa.gov/docs/code/422.7.pdf",
    )
    defined_for = StateCode.IA

    def formula(tax_unit, period, parameters):
        fed_taxable_income = tax_unit("taxable_income", period)
        modifications = tax_unit(
            "ia_taxable_income_modifications_consolidated", period
        )
        # Modifications include additions and subtractions
        return max_(0, fed_taxable_income + modifications)



from numpy import clip
from policyengine_us.model_api import *


class capital_gains_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maximum income tax after capital gains tax"
    unit = USD
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        net_cg = tax_unit("net_capital_gain", period)
        taxable_income = tax_unit("taxable_income", period)
        adjusted_net_cg = min_(
            tax_unit("adjusted_net_capital_gain", period),
            taxable_income,
        )  # ANCG is referred to in all cases as ANCG or taxable income if less.

        cg = parameters(period).gov.irs.capital_gains

        excluded_cg = tax_unit(
            "capital_gains_excluded_from_taxable_income", period
        )
        non_cg_taxable_income = max_(0, taxable_income - excluded_cg)
        income_less_ancg = max_(0, taxable_income - adjusted_net_cg)

        filing_status = tax_unit("filing_status", period)

        first_threshold = cg.thresholds["1"][filing_status]
        second_threshold = cg.thresholds["2"][filing_status]

        income_ordinarily_under_second_rate = clip(
            taxable_income, 0, first_threshold
        )
        cg_in_first_bracket = max_(
            0, income_ordinarily_under_second_rate - income_less_ancg
        )

        income_ordinarily_under_third_rate = clip(
            taxable_income, 0, second_threshold
        )
        cg_in_second_bracket = min_(
            max_(0, adjusted_net_cg - cg_in_first_bracket),
            max_(
                0,
                income_ordinarily_under_third_rate
                - (non_cg_taxable_income + cg_in_first_bracket),
            ),
        )

        cg_in_third_bracket = max_(
            adjusted_net_cg - cg_in_first_bracket - cg_in_second_bracket,
            0,
        )

        main_cg_tax = (
            cg_in_first_bracket * cg.rates["1"]
            + cg_in_second_bracket * cg.rates["2"]
            + cg_in_third_bracket * cg.rates["3"]
        )

        unrecaptured_s_1250_gain = tax_unit(
            "unrecaptured_section_1250_gain", period
        )
        qualified_dividends = add(
            tax_unit, period, ["qualified_dividend_income"]
        )
        max_taxable_unrecaptured_gain = min_(
            unrecaptured_s_1250_gain,
            max_(0, net_cg - qualified_dividends),
        )
        unrecaptured_gain_deduction = max_(
            non_cg_taxable_income + net_cg - taxable_income,
            0,
        )
        taxable_unrecaptured_gain = max_(
            max_taxable_unrecaptured_gain - unrecaptured_gain_deduction,
            0,
        )

        unrecaptured_gain_tax = (
            cg.unrecaptured_s_1250_rate * taxable_unrecaptured_gain
        )

        remaining_cg_tax = (
            tax_unit("capital_gains_28_percent_rate_gain", period)
            * cg.other_cg_rate
        )

        return main_cg_tax + unrecaptured_gain_tax + remaining_cg_tax



from policyengine_us.model_api import *


class deep_poverty_line(Variable):
    value_type = float
    entity = SPMUnit
    label = "deep poverty line"
    unit = USD
    documentation = "Income threshold below which a household is considered to be in deep poverty."
    definition_period = YEAR

    def formula(spm_unit, period, parameters):
        return spm_unit("poverty_line", period) / 2



from policyengine_us.model_api import *


class is_usda_elderly(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    documentation = "Is elderly per USDA guidelines"
    label = "USDA elderly"

    def formula(person, period, parameters):
        elderly_age_threshold = parameters(
            period
        ).gov.usda.elderly_age_threshold
        return person("age", period) >= elderly_age_threshold



from policyengine_us.model_api import *


class va_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Virginia standard deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://law.lis.virginia.gov/vacodefull/title58.1/chapter3/article2/"
    )
    defined_for = StateCode.VA

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.va.tax.income.deductions
        filing_status = tax_unit("filing_status", period)
        return p.standard[filing_status]



from policyengine_us.model_api import *


class ga_itemizer_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Georgia itemizer tax credit"
    unit = USD
    definition_period = YEAR
    reference = "https://law.justia.com/codes/georgia/2022/title-48/chapter-7/article-2/section-48-7-29-23/"
    defined_for = StateCode.GA

    def formula(tax_unit, period, parameters):
        # Full-year and part-year residents who itemize their deductions
        # are entitled to a credit up to $300 per taxpayer
        itemizes = tax_unit("tax_unit_itemizes", period)
        taxpayer_count = tax_unit("head_spouse_count", period)
        p = parameters(period).gov.states.ga.tax.income.credits.itemizer
        return where(itemizes, p.amount * taxpayer_count, 0)



from policyengine_us.model_api import *


class il_tanf_countable_earned_income_for_grant_calculation(Variable):
    value_type = float
    entity = SPMUnit
    label = "Illinois Temporary Assistance for Needy Families (TANF) countable earned income for grant calculation"
    unit = USD
    definition_period = MONTH
    reference = "https://www.dhs.state.il.us/page.aspx?item=15864"
    defined_for = StateCode.IL

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.il.dhs.tanf.income.disregard

        countable_gross_earned_income = spm_unit(
            "il_tanf_countable_gross_earned_income", period
        )
        childcare_deduction = spm_unit("il_tanf_childcare_deduction", period)
        adjusted_earned_income = max_(
            countable_gross_earned_income - childcare_deduction, 0
        )
        adjusted_earned_income_disregard = p.rate * adjusted_earned_income
        return max_(
            adjusted_earned_income - adjusted_earned_income_disregard, 0
        )



from policyengine_us.model_api import *


class taxable_earnings_for_social_security(Variable):
    value_type = float
    entity = Person
    label = "Taxable gross earnings for OASDI FICA"
    definition_period = YEAR
    unit = USD

    def formula(person, period, parameters):
        p = parameters(period).gov.irs.payroll.social_security
        return min_(p.cap, person("payroll_tax_gross_wages", period))



from policyengine_us.model_api import *


class capped_energy_efficient_insulation_credit(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Capped energy efficient insulation credit"
    documentation = "Capped credit on energy-efficient insulation material"
    unit = USD
    reference = "https://www.democrats.senate.gov/imo/media/doc/inflation_reduction_act_of_2022.pdf#page=339"

    def formula(tax_unit, period, parameters):
        pre_rebate_expenditure = tax_unit(
            "energy_efficient_insulation_expenditures", period
        )
        p = parameters(
            period
        ).gov.irs.credits.energy_efficient_home_improvement
        # NB: We assume that the credit is based on after-rebate expenditures,
        # where rebates are per-item before the total rebate cap is applied.
        # We also assume all rebates for insulation, air sealing, and ventilation
        # are for insulation. Otherwise would require allocating the rebate
        # across subcategories.
        rebate = tax_unit(
            "capped_insulation_air_sealing_ventilation_rebate", period
        )
        # Cap at zero in case they have a rebate from air sealing and ventilation
        # but no insulation expenditures.
        post_rebate_expenditure = max_(0, pre_rebate_expenditure - rebate)
        rate = p.rates.improvements
        uncapped = post_rebate_expenditure * rate
        return min_(uncapped, p.cap.annual.insulation_material)



from policyengine_us.model_api import *


class pr_education_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Puerto Rico education contribution deduction"
    unit = USD
    definition_period = YEAR
    reference = "https://law.justia.com/codes/puerto-rico/title-thirteen/subtitle-17/part-ii/chapter-1005/subchapter-c/30135/"  # (8)
    defined_for = StateCode.PR

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.territories.pr.tax.income.taxable_income.deductions.education
        contributions = add(tax_unit, period, ["investment_in_529_plan"])
        count = tax_unit("pr_education_deduction_beneficiary_count", period)
        maximum_contribution = p.max * count
        return min_(maximum_contribution, contributions)



from policyengine_us.model_api import *


class is_snap_ineligible_student(Variable):
    value_type = bool
    entity = Person
    label = "Is an ineligible student for SNAP"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/7/2015#e"

    def formula(person, period, parameters):
        # Base rule: Students enrolled at least half-time in higher education
        # are ineligible (K-12 students are not affected by this rule)
        is_higher_ed_student = person("is_snap_higher_ed_student", period)

        # Eight statutory exceptions that make students eligible:

        # Exception 1: Under 18 or age 50 or older
        age = person("age", period)
        p = parameters(period).gov.usda.snap.student
        meets_age_exception = p.age_threshold.calc(age)

        # Exception 2: Not physically or mentally fit (disabled)
        meets_disability_exception = person("is_disabled", period)

        # Exception 3: Assignment through workforce/employment programs
        # (WIOA, career/technical ed, remedial/basic education)
        # Not modeled

        # Exception 4: Employed at least 20 hours per week or work-study
        meets_work_hours_exception = person(
            "meets_snap_work_exception", period
        )

        # Exception 5: Parent with responsibility for dependent child under 6,
        # or child 6-11 when adequate child care is not available
        # Exception 8: Single parent enrolled full-time with responsibility
        # for dependent child under 12
        # (Exceptions 5 and 8 are implemented together)
        meets_parent_exception = person("meets_snap_parent_exception", period)

        # Exception 6: Receiving TANF benefits (part A of title IV)
        tanf = person("tanf_person", period)
        receives_tanf = tanf > 0

        # Exception 7: Enrolled as result of participation in work incentive
        # program under title IV (TANF work programs) or successor programs
        # Not modeled

        # Student is INELIGIBLE if they are a higher ed student AND
        # they do NOT meet ANY of the eight exceptions
        meets_any_exception = (
            meets_age_exception
            | meets_disability_exception
            | meets_work_hours_exception
            | meets_parent_exception
            | receives_tanf
        )

        return is_higher_ed_student & ~meets_any_exception



from policyengine_us.model_api import *


class ar_personal_credits_base(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arkansas base personal credits"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2021_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf"
        "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_FullYearResidentIndividualIncomeTaxReturn.pdf#page=1"
        "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf#page=12"
    )
    defined_for = StateCode.AR

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        # Only head and spouse are eligible for the personal credit amounts
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        p = parameters(period).gov.states.ar.tax.income.credits.personal
        aged = person("age", period) >= p.age_threshold
        # Arkansas provides an additional "aged special" credit for people who
        # do not receive retirement or disability benefit exemptions.
        receives_retirement_or_disability_exemption = (
            person(
                "ar_retirement_or_disability_benefits_exemption_person", period
            )
            > 0
        )
        aged_special = aged & ~receives_retirement_or_disability_exemption
        # Blind filers get an additional personal tax credit amount
        blind = person("is_blind", period)
        # Deaf filers get an additional personal tax credit amount
        deaf = person("is_deaf", period)
        # surviving spouse and head of household filers receive an additional credit amount
        filing_status = tax_unit("filing_status", period)
        statuses = filing_status.possible_values
        surviving_spouse = filing_status == statuses.SURVIVING_SPOUSE
        hoh = filing_status == statuses.HEAD_OF_HOUSEHOLD
        filing_status_eligible = surviving_spouse | hoh

        personal_credit_count = tax_unit.sum(
            head_or_spouse
            * (
                1
                + aged.astype(int)
                + blind.astype(int)
                + deaf.astype(int)
                + aged_special.astype(int)
            )
        ) + filing_status_eligible.astype(int)
        return personal_credit_count * p.amount.base



from policyengine_us.model_api import *


class wi_childcare_expense_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Wisconsin childcare expense credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.wi.gov/TaxForms2022/2022-Form1f.pdf#page=2"
        "https://www.revenue.wi.gov/TaxForms2022/2022-Form1-Inst.pdf#page=17"
        "https://docs.legis.wisconsin.gov/misc/lfb/informational_papers/january_2023/0002_individual_income_tax_informational_paper_2.pdf"
    )
    defined_for = StateCode.WI

    def formula(tax_unit, period, parameters):
        # Wisconsin matches the potential federal credit
        us_cdcc = tax_unit("cdcc_potential", period)
        p = parameters(period).gov.states.wi.tax.income
        return us_cdcc * p.credits.childcare_expense.fraction



from policyengine_us.model_api import *


class nc_income_tax_before_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "North Carolina income tax before credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NC

    def formula(tax_unit, period, parameters):
        income = tax_unit("nc_taxable_income", period)
        p = parameters(period).gov.states.nc.tax.income
        return max_(0, income) * p.rate



from policyengine_us.model_api import *


class az_tanf_eligible_child(Variable):
    value_type = bool
    entity = Person
    label = "Eligible child for the Arizona Cash Assistance"
    definition_period = YEAR
    defined_for = StateCode.AZ

    def formula(person, period, parameters):
        # Get the age of the person
        age = person("age", period)
        # Determine whether they are a student
        student = person("is_full_time_student", period)
        # Determine the age thresholds
        p = parameters(period).gov.states.az.hhs.tanf.eligibility.age_threshold
        age_threshold = where(student, p.student, p.non_student)
        return age < age_threshold



from policyengine_us.model_api import *


class ma_eaedc_earned_income_after_disregard_person(Variable):
    value_type = float
    entity = Person
    label = "Massachusetts EAEDC earned income after disregard for each person"
    unit = USD
    definition_period = MONTH
    defined_for = StateCode.MA
    reference = "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-704-500"  # (B) step 2

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.ma.dta.tcap.eaedc.deductions.income_disregard
        gross_income = person("ma_tcap_gross_earned_income", period)
        work_related_expenses_deduction = person(
            "ma_tafdc_work_related_expense_deduction", period
        )
        income_after_work_related_expenses_deduction = max_(
            gross_income - work_related_expenses_deduction, 0
        )
        income_after_flat_disregard = max_(
            income_after_work_related_expenses_deduction - p.flat,
            0,
        )
        percentage_disregard = income_after_flat_disregard * p.percentage.rate
        if period.start.month <= p.percentage.months:
            return max_(income_after_flat_disregard - percentage_disregard, 0)
        # A percentage disregard is applied for the first 4 months in addition to the continuous flat disregard.
        return income_after_flat_disregard



from policyengine_us.model_api import *


class investment_in_529_plan_indv(Variable):
    value_type = float
    entity = Person
    label = "Individual 529 plan investment amounts"
    unit = USD
    documentation = (
        "Amount invested in a 529 savings plan for each contributor."
    )
    definition_period = YEAR



from policyengine_us.model_api import *


class assessed_property_value(Variable):
    value_type = float
    entity = Person
    label = "Assessed property value"
    unit = USD
    documentation = "Total assessed value of property owned by this person."
    definition_period = YEAR



import pytest

DATASETS = [
    "hf://policyengine/policyengine-us-data/cps_2023.h5",
    "hf://policyengine/policyengine-us-data/enhanced_cps_2024.h5",
]

YEARS = list(range(2024, 2026))


@pytest.mark.parametrize("dataset", DATASETS)
@pytest.mark.parametrize("year", YEARS)
def test_microsim_runs(dataset: str, year: int):
    import numpy as np
    from policyengine_us import Microsimulation

    sim = Microsimulation(dataset=dataset)
    sim.subsample(1_000)
    hnet = sim.calc("household_net_income", period=year)
    assert not hnet.isna().any(), "Some households have NaN net income."
    # Deciles are 1-10, with -1 for negative income.
    DECILES = [
        "household_income_decile",
        "spm_unit_income_decile",
        "income_decile",
    ]
    for decile_var in DECILES:
        decile = sim.calc(decile_var)
        assert np.all(decile >= -1) and np.all(
            decile <= 10
        ), f"{decile_var} out of bounds."

    # Check that the microsim calculates important variables as nonzero in current year.
    for var in ["employment_income", "self_employment_income"]:
        assert sim.calc(var, period=2024).sum() > 0, f"{var} is zero in 2024."



from policyengine_us.model_api import *


class wa_tanf_countable_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "Washington TANF countable income"
    unit = USD
    definition_period = MONTH
    reference = "https://app.leg.wa.gov/wac/default.aspx?cite=388-450-0165"
    defined_for = StateCode.WA

    adds = [
        "wa_tanf_countable_earned_income",
        "tanf_gross_unearned_income",
    ]



from policyengine_us.model_api import *


class ut_income_tax_before_refundable_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "Utah income tax before refundable credits"
    unit = USD
    definition_period = YEAR
    documentation = "Form TC-40, line 32"
    defined_for = StateCode.UT

    def formula(tax_unit, period, parameters):
        income = tax_unit(
            "ut_income_tax_before_non_refundable_credits", period
        )
        credits = tax_unit("ut_non_refundable_credits", period)
        return max_(income - credits, 0)



from policyengine_us.model_api import *


class is_premium_free_part_a(Variable):
    value_type = bool
    entity = Person
    label = "Qualifies for premium-free Part A"
    documentation = (
        "Whether the person qualifies for premium-free Medicare Part A "
        "based on having 40 or more quarters of Medicare-covered employment. "
        "Approximately 99% of Medicare beneficiaries qualify for premium-free Part A."
    )
    definition_period = YEAR
    reference = (
        "https://www.medicare.gov/basics/costs/medicare-costs/part-a-costs"
    )
    defined_for = "is_medicare_eligible"

    def formula(person, period, parameters):
        p = parameters(period).gov.hhs.medicare.part_a
        quarters = person("medicare_quarters_of_coverage", period)
        return quarters >= p.premium_free_quarters_threshold



from policyengine_us.model_api import *


class pa_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Pennsylvania Working Pennsylvanians Tax Credit"
    defined_for = StateCode.PA
    unit = USD
    definition_period = YEAR
    reference = "https://www.palegis.us/legislation/bills/text/PDF/2025/0/HB0416/PN2576#page=50"  # Article XVI-W.2

    def formula(tax_unit, period, parameters):
        federal_eitc = tax_unit("eitc", period)
        p = parameters(period).gov.states.pa.tax.income.credits.eitc
        return federal_eitc * p.match



from policyengine_us.model_api import *


class hhs_smi(Variable):
    value_type = float
    entity = SPMUnit
    label = "State Median Income (HHS)"
    documentation = "SPM unit's median income as defined by the Department of Health and Human Services, based on their state and size"
    definition_period = YEAR
    unit = USD

    def formula(spm_unit, period, parameters):
        size = spm_unit("spm_unit_size", period)
        state = spm_unit.household("state_code_str", period)
        four_person_smi = parameters(period).gov.hhs.smi.amount[state]
        adjustment_mapping = parameters(
            period
        ).gov.hhs.smi.household_size_adjustment
        first_person_rate = adjustment_mapping.first_person
        second_to_sixth_additional_rate = (
            adjustment_mapping.second_to_sixth_person
        )
        seven_or_more_additional_rate = adjustment_mapping.additional_person
        size_adjustment = (
            first_person_rate
            + second_to_sixth_additional_rate * (min_(size, 6) - 1)
            + seven_or_more_additional_rate * max_(size - 6, 0)
        )
        return four_person_smi * size_adjustment



from policyengine_us.model_api import *


class dc_liheap_income_level(Variable):
    value_type = int
    entity = SPMUnit
    label = "Income level for DC LIHEAP payment"
    definition_period = YEAR
    reference = "https://doee.dc.gov/sites/default/files/dc/sites/doee/service_content/attachments/DOEE%20FY24%20LIHEAP_REGULAR_Benefits_Table-Matrix.pdf"
    defined_for = StateCode.DC

    def formula(spm_unit, period, parameters):
        income = add(spm_unit, period, ["irs_gross_income"])
        p = parameters(period).gov.states.dc.doee.liheap

        levels = [i * p.income_level_increment for i in range(1, 10)]

        return select(
            [income <= level for level in levels],
            list(range(1, 10)),
            default=10,
        )



from policyengine_us.model_api import *


class ms_itemized_deductions_unit(Variable):
    value_type = float
    entity = TaxUnit
    label = "Mississippi itemized deductions"
    unit = USD
    definition_period = YEAR

    reference = (
        "https://www.dor.ms.gov/sites/default/files/Forms/Individual/80100221.pdf#page=15"
        "https://www.dor.ms.gov/sites/default/files/Forms/Individual/80108228.pdf, # Line 7"
        "https://casetext.com/statute/mississippi-code-1972/title-27-taxation-and-finance/chapter-7-income-tax-and-withholding/article-1-income-tax/section-27-7-17-deductions-allowed?__cf_chl_rt_tk=8Kelu8kHpIXTp_FnAJLHvqa7rtrZYE1U.NAeBM8L.Nc-1692990420-0-gaNycGzNEmU"
    )
    defined_for = StateCode.MS

    adds = [
        "itemized_deductions_less_salt",
        "misc_deduction",
        "ms_real_estate_tax_deduction",
    ]

    # Mississippi allows itemized deductions for gaming establishments and gender transition procedures
    # which are currently not modeled



from policyengine_us.model_api import *


class w2_wages_from_qualified_business(Variable):
    value_type = float
    entity = Person
    label = "W2 wages"
    unit = USD
    documentation = "Share of wages paid by this person to employees as part of a pass-through qualified business or trade."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/199A#b_4"



from policyengine_us.model_api import *


class il_hbwd_disability_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Illinois Health Benefits for Workers with Disabilities disability eligible"
    definition_period = MONTH
    reference = (
        "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-120.510",
        "https://hfs.illinois.gov/medicalprograms/hbwd/eligibility.html",
    )
    defined_for = StateCode.IL

    def formula(person, period, parameters):
        # Must meet Social Security Administration's definition of disability
        is_ssi_disabled = person("is_ssi_disabled", period)
        # Or currently receiving SSDI
        receives_ssdi = person("social_security_disability", period) > 0
        return is_ssi_disabled | receives_ssdi



from .ctc_older_child_supplement import (
    create_ctc_older_child_supplement_reform,
)
from .ctc_additional_bracket import (
    create_ctc_additional_bracket_reform,
)
from .ctc_per_child_phase_out import (
    create_ctc_per_child_phase_out_reform,
)
from .ctc_per_child_phase_in import (
    create_ctc_per_child_phase_in_reform,
)
from .ctc_minimum_refundable_amount import (
    create_ctc_minimum_refundable_amount_reform,
)



from policyengine_us.model_api import *


class mt_capital_gains_tax_indiv(Variable):
    value_type = float
    entity = Person
    label = "Montana net long-term capital gains tax when married couples file separately"
    unit = USD
    definition_period = YEAR
    reference = "https://mtrevenue.gov/wp-content/uploads/dlm_uploads/2023/12/Form_2_2023_Instructions.pdf#page=6"  # Net Long-Term Capital Gains Tax Table
    defined_for = StateCode.MT

    def formula(person, period, parameters):
        p = parameters(period).gov.states.mt.tax.income.main.capital_gains
        # the tax for capital gains comes into effect after 2024
        if p.in_effect:
            capital_gains = person("long_term_capital_gains", period)
            filing_status = person.tax_unit(
                "state_filing_status_if_married_filing_separately_on_same_return",
                period,
            )
            applicable_threshold = person(
                "mt_capital_gains_tax_applicable_threshold_indiv", period
            )
            status = filing_status.possible_values

            lower_rate = select(
                [
                    filing_status == status.SINGLE,
                    filing_status == status.SEPARATE,
                    filing_status == status.SURVIVING_SPOUSE,
                    filing_status == status.HEAD_OF_HOUSEHOLD,
                ],
                [
                    p.rates.single.amounts[0],
                    p.rates.separate.amounts[0],
                    p.rates.surviving_spouse.amounts[0],
                    p.rates.head_of_household.amounts[0],
                ],
            )
            higher_rate = select(
                [
                    filing_status == status.SINGLE,
                    filing_status == status.SEPARATE,
                    filing_status == status.SURVIVING_SPOUSE,
                    filing_status == status.HEAD_OF_HOUSEHOLD,
                ],
                [
                    p.rates.single.amounts[-1],
                    p.rates.separate.amounts[-1],
                    p.rates.surviving_spouse.amounts[-1],
                    p.rates.head_of_household.amounts[-1],
                ],
            )
            # Calculate taxes
            capital_gains_below_threshold = min_(
                applicable_threshold, capital_gains
            )
            capital_gains_above_threshold = max_(
                capital_gains - applicable_threshold, 0
            )

            lower_capital_gains_tax = (
                capital_gains_below_threshold * lower_rate
            )
            higher_capital_gains_tax = (
                capital_gains_above_threshold * higher_rate
            )

            return lower_capital_gains_tax + higher_capital_gains_tax

        return 0



from policyengine_us.model_api import *
from policyengine_us.variables.gov.ssa.ssi.eligibility.income._apply_ssi_exclusions import (
    _apply_ssi_exclusions,
)


class ssi_unearned_income_deemed_from_ineligible_parent(Variable):
    value_type = float
    entity = Person
    label = "SSI unearned income (deemed from ineligible parent)"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/cfr/text/20/416.1165"

    def formula(person, period, parameters):
        eligible_child = person("is_ssi_aged_blind_disabled", period) & person(
            "is_child", period
        )
        ineligible_parent = person("is_ssi_ineligible_parent", period)
        tax_unit = person.tax_unit

        prereduction_parental_earned_income = eligible_child * tax_unit.sum(
            person("ssi_earned_income", period) * ineligible_parent
        )
        blind_disabled_working_student_income = person(
            "ssi_blind_or_disabled_working_student_exclusion", period
        )
        parental_earned_income = max_(
            prereduction_parental_earned_income
            - blind_disabled_working_student_income,
            0,
        )

        parental_unearned_income = eligible_child * tax_unit.sum(
            person("ssi_unearned_income", period) * ineligible_parent
        )

        child_allocations = eligible_child * add(
            tax_unit, period, ["ssi_ineligible_child_allocation"]
        )
        parental_allocations = eligible_child * add(
            tax_unit, period, ["ssi_ineligible_parent_allocation"]
        )

        # To understand this ordering of operations, see Example 4 of
        # https://www.law.cornell.edu/cfr/text/20/416.1165#h.

        parental_unearned_income -= child_allocations
        remaining_child_allocations = max_(0, -parental_unearned_income)
        parental_unearned_income = max_(0, parental_unearned_income)
        parental_earned_income = max_(
            0, parental_earned_income - remaining_child_allocations
        )

        net_parental_deemed_income = _apply_ssi_exclusions(
            parental_earned_income,
            parental_unearned_income,
            parameters,
            period,
        )

        net_parental_deemed_income = max_(
            0, net_parental_deemed_income - parental_allocations
        )

        count_eligible_children = tax_unit.sum(eligible_child)
        # avoid array divide-by-zero warnings by not using where() function
        # see the following GitHub issue for more details:
        # https://github.com/PolicyEngine/policyengine-us/issues/2494
        income = np.zeros_like(count_eligible_children)
        mask = count_eligible_children > 0
        income[mask] = (
            net_parental_deemed_income[mask] / count_eligible_children[mask]
        )
        return income



from policyengine_us.model_api import *


class person_family_id(Variable):
    value_type = int
    entity = Person
    label = "Unique reference for the family of this person"
    definition_period = YEAR



from policyengine_us.model_api import *


class SchoolMealTier(Enum):
    FREE = "Free"
    REDUCED = "Reduced price"
    PAID = "Paid"


class school_meal_tier(Variable):
    value_type = Enum
    label = "School meal tier"
    possible_values = SchoolMealTier
    default_value = SchoolMealTier.PAID
    entity = SPMUnit
    definition_period = YEAR
    documentation = "SPM unit's school meal program tier"

    def formula(spm_unit, period, parameters):
        fpg_ratio = spm_unit("school_meal_fpg_ratio", period)
        p_income_limit = parameters(period).gov.usda.school_meals.income.limit
        # Categorical eligibility provides free school meals.
        categorical_eligibility = spm_unit(
            "meets_school_meal_categorical_eligibility", period
        )
        return select(
            [
                (fpg_ratio <= p_income_limit.FREE) | categorical_eligibility,
                fpg_ratio <= p_income_limit.REDUCED,
                True,
            ],
            [SchoolMealTier.FREE, SchoolMealTier.REDUCED, SchoolMealTier.PAID],
        )



from policyengine_us.model_api import *


class il_hbwd(Variable):
    value_type = float
    entity = SPMUnit
    label = "Illinois Health Benefits for Workers with Disabilities"
    unit = USD
    definition_period = MONTH
    reference = (
        "https://hfs.illinois.gov/medicalprograms/hbwd.html",
        "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-120.510",
        "https://hfs.illinois.gov/medicalprograms/hbwd/about.html",
    )
    defined_for = StateCode.IL

    adds = ["il_hbwd_person"]



from policyengine_us.model_api import *


class is_federal_work_study_participant(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    label = "Participating in Federal Work Study"
    reference = "https://www.law.cornell.edu/cfr/text/34/part-675"



from policyengine_us.model_api import *


class vt_subtractions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Vermont subtractions"
    unit = USD
    documentation = "Subtractions from Vermont adjusted gross income"
    definition_period = YEAR
    defined_for = StateCode.VT
    reference = (
        "https://tax.vermont.gov/sites/tax/files/documents/IN-112-2022.pdf#page=1",  # PART 1 SUBTRACTIONS TO FEDERAL ADJUSTED GROSS INCOME
        "https://legislature.vermont.gov/statutes/section/32/151/05811",  # Titl. 32 V.S.A. § 5811(21)(B)(i), (C)(iv), (B)(vi), (B)(ii), (B)(iv)
        "https://tax.vermont.gov/sites/tax/files/documents/IN-112%20Instr-2022.pdf",
    )

    # Get parameter list
    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.vt.tax.income.agi
        total_subtractions = add(tax_unit, period, p.subtractions)
        # Prevent negative subtractions from acting as additions
        return max_(0, total_subtractions)



from policyengine_us.model_api import *


class is_tafdc_related_to_head_or_spouse(Variable):
    value_type = bool
    entity = Person
    label = (
        "Person is related to the head or spouse under the TAFDC regulations"
    )
    definition_period = YEAR
    reference = (
        "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-703-310"
    )
    default_value = True



from policyengine_us.model_api import *


class early_head_start(Variable):
    value_type = float
    entity = Person
    label = "Early Head Start"
    unit = USD
    definition_period = YEAR
    defined_for = "is_early_head_start_eligible"
    reference = "https://headstart.gov/program-data/article/head-start-program-facts-fiscal-year-2022"

    def formula(person, period, parameters):
        p = parameters(period).gov.hhs.head_start.early_head_start
        state = person.household("state_code_str", period)
        spending = p.spending[state]
        enrollment = p.enrollment[state]
        mask = enrollment > 0
        result = np.zeros_like(p.spending[state])
        result[mask] = spending[mask] / enrollment[mask]
        return result



from policyengine_us.model_api import *


class mi_personal_exemptions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Michigan personal and stillborn exemptions"
    defined_for = StateCode.MI
    unit = USD
    definition_period = YEAR
    reference = (
        "http://legislature.mi.gov/doc.aspx?mcl-206-30",
        "https://www.legislature.mi.gov/Publications/TaxpayerGuide.pdf",
    )

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.mi.tax.income.exemptions

        # Personal Exemptions & Stillborn Exemptions
        exemptions = add(
            tax_unit, period, ["tax_unit_size", "tax_unit_stillborn_children"]
        )
        return exemptions * p.personal



from policyengine_us.model_api import *


class spm_unit_total_ccdf_copay(Variable):
    value_type = float
    entity = SPMUnit
    definition_period = YEAR
    label = "SPM unit total CCDF copay"
    unit = USD
    reference = "https://www.ocfs.ny.gov/programs/childcare/stateplan/assets/2022-plan/FFY2022-2024-CCDF-Plan.pdf#page=107"

    def formula(spm_unit, period, parameters):
        # Get family income and federal poverty line
        income = spm_unit("ccdf_income", period)
        fpl = spm_unit("spm_unit_fpg", period)
        # Get county and copay percent parameter by county
        county = spm_unit.household("county_str", period)
        p_copay_percent = np.ones_like(county)
        copays = parameters(period).gov.hhs.ccdf.copay_percent
        valid_county = np.isin(county, np.array(list(copays._children)))
        if valid_county.sum() > 0:
            p_copay_percent[valid_county] = parameters(
                period
            ).gov.hhs.ccdf.copay_percent[county[valid_county]]
        income_exceeding_fpl = max_(income - fpl, 0)
        # There is a minimum family fee of $1 per week
        return max_(income_exceeding_fpl * p_copay_percent, 52)



from policyengine_us.model_api import *


class snap_deductions(Variable):
    value_type = float
    entity = SPMUnit
    label = "SNAP income deductions"
    unit = USD
    documentation = "Deductions made from gross income for SNAP benefits"
    definition_period = MONTH
    reference = "https://www.law.cornell.edu/uscode/text/7/2014#e"

    adds = "gov.usda.snap.income.deductions.allowed"



from policyengine_us.model_api import *


class tx_ccs_asset_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Texas CCS asset eligible"
    definition_period = MONTH
    reference = "https://www.law.cornell.edu/regulations/texas/40-Tex-Admin-Code-SS-809-41"
    defined_for = StateCode.TX

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.tx.twc.ccs
        assets = spm_unit("spm_unit_assets", period.this_year)
        asset_limit = p.assets.limit
        return assets <= asset_limit



from policyengine_us.model_api import *


class dc_tanf_countable_income(Variable):
    value_type = float
    entity = SPMUnit
    label = (
        "DC Temporary Assistance for Needy Families (TANF) countable income"
    )
    unit = USD
    definition_period = MONTH
    defined_for = StateCode.DC

    def formula(spm_unit, period, parameters):
        income_sources = add(
            spm_unit,
            period,
            [
                "dc_tanf_countable_earned_income",
                "dc_tanf_countable_unearned_income",
            ],
        )
        # Exclude income from GAC, this only applies when the child in household is unrelated to
        # the head or spouse
        gac_countable_income = spm_unit("dc_gac_countable_income", period)
        childcare_deduction = spm_unit("dc_tanf_childcare_deduction", period)
        return max_(
            income_sources - gac_countable_income - childcare_deduction, 0
        )



from policyengine_us.model_api import *


class dc_files_separately(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Married couple files separately on DC tax return"
    definition_period = YEAR
    reference = ()
    defined_for = StateCode.DC

    def formula(tax_unit, period, parameters):
        itax_indiv = tax_unit("dc_income_tax_before_credits_indiv", period)
        itax_joint = tax_unit("dc_income_tax_before_credits_joint", period)
        if parameters(period).gov.states.dc.tax.income.joint_separately_option:
            return itax_indiv < itax_joint
        return False



from policyengine_us.model_api import *


class ny_additional_ctc(Variable):
    value_type = float
    entity = TaxUnit
    label = "New York additional Empire State Child Credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.nysenate.gov/legislation/laws/TAX/606"  # (c-1)(4)(A)
    )
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        agi = tax_unit("adjusted_gross_income", period)
        p = parameters(period).gov.states.ny.tax.income.credits.ctc.additional
        match = p.amount.calc(agi)
        base_ctc = tax_unit("ny_ctc", period)
        credit_amount = base_ctc * match
        credit_above_min = credit_amount >= p.min_credit_value
        return credit_above_min * credit_amount



from policyengine_us.model_api import *


class employment_income_last_year(Variable):
    value_type = float
    entity = Person
    label = "employment income last year"
    documentation = (
        "Wages and salaries in prior year, including tips and commissions."
    )
    unit = USD
    definition_period = YEAR
    uprating = "calibration.gov.irs.soi.employment_income"

    def formula_2024(person, period, parameters):
        employment_income_target = (
            parameters.calibration.gov.irs.soi.employment_income
        )
        value_last_year = employment_income_target(period.last_year)
        value_year_before_last = employment_income_target(
            period.last_year.last_year
        )
        values = person("employment_income_last_year", period.last_year)
        uprating_factor = value_last_year / value_year_before_last
        return values * uprating_factor



from policyengine_us.model_api import *
from policyengine_core.simulations import Simulation


class is_ssi_aged_blind_disabled(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    documentation = "Indicates whether a person is aged, blind, or disabled for the Supplemental Security Income program"
    label = "SSI aged, blind, or disabled"
    reference = "https://www.law.cornell.edu/uscode/text/42/1382c#a_1"

    def formula(person, period, parameters):
        simulation: Simulation = person.simulation
        return any_(
            person, period, ["is_ssi_aged", "is_blind", "is_ssi_disabled"]
        )



from policyengine_us.model_api import *


class id_ctc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Idaho Child Tax Credit"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.ID

    def formula(tax_unit, period, parameters):
        # Get relevant parameter subtree.
        p = parameters(period).gov.states.id.tax.income.credits.ctc
        # "the term "qualifying child" has the meaning as defined in section 24(c) of the Internal Revenue Code."
        # IRC 24(c) defines eligible children under the federal CTC.
        eligible_children = tax_unit("ctc_qualifying_children", period)
        # Multiply by the amount per child.
        return eligible_children * p.amount



from policyengine_us.model_api import *


class or_wfhdc_income_category(Variable):
    value_type = int
    entity = TaxUnit
    label = "Oregon working family household and dependent care credit percentage table row letter"
    unit = "/1"
    definition_period = YEAR
    defined_for = "or_wfhdc_eligible"
    reference = "https://www.oregon.gov/dor/forms/FormsPubs/publication-or-wfhdc-tb_101-458_2021.pdf#page=1"

    def formula(tax_unit, period, parameters):
        # Get the household income, considered the larger of Federal and Oregon AGI.
        household_income = tax_unit("or_wfhdc_household_income", period)
        # The credit percentage is based on the percentage of the tax unit fpg
        fpg = tax_unit("tax_unit_fpg", period)
        fpg_rate = household_income / fpg
        # The rate can not drop below 0%.
        floored_fpg_rate = max_(fpg_rate, 0)
        # The rate can not exceed 300%.
        return np.ceil(floored_fpg_rate * 10)



from policyengine_us.model_api import *
from microdf import MicroSeries


class spm_unit_income_decile(Variable):
    value_type = int
    entity = SPMUnit
    label = "Income decile"
    documentation = "The income decile of the SPM unit, person-weighted and using OECD-equivalised net income"
    definition_period = YEAR

    def formula(spm_unit, period, parameters):
        income = spm_unit("spm_unit_oecd_equiv_net_income", period)
        weights = spm_unit("spm_unit_weight", period) * spm_unit.nb_persons()
        return MicroSeries(income, weights=weights).decile_rank()



from policyengine_us.model_api import *


class ma_tafdc_countable_earned_income(Variable):
    value_type = float
    unit = USD
    entity = Person
    label = "Massachusetts Temporary Assistance for Families with Dependent Children (TAFDC) countable earned income"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/regulations/massachusetts/106-CMR-704-281"  # (A)&(B)
    defined_for = StateCode.MA

    def formula(person, period, parameters):
        # This variable is used for the TAFDC benefit amount calculation
        # The variable is computed yearly to account for the 100% disregard for the first 6 months of the year
        p = parameters(
            period
        ).gov.states.ma.dta.tcap.tafdc.earned_income_disregard
        total_earned_income = person("ma_tcap_gross_earned_income", period)
        # fully disregard for 6 months
        full_disregard_eligible = person(
            "ma_tafdc_full_earned_income_disregard_eligible", period
        )
        full_disregard = (
            total_earned_income
            - total_earned_income * p.full_disregard.percentage
        ) * p.full_disregard.applicable_months
        remaining_months = MONTHS_IN_YEAR - p.full_disregard.applicable_months
        # The partially disregarded earned income is either applied for 6 months if the unit received
        # the full disregard or for the entire year if the unit did not receive the full disregard.
        partially_disregarded_income = (
            person("ma_tafdc_partially_disregarded_earned_income", period)
            / MONTHS_IN_YEAR
        )
        last_6_months = partially_disregarded_income * remaining_months
        first_6_months = where(
            full_disregard_eligible,
            full_disregard,
            partially_disregarded_income * p.full_disregard.applicable_months,
        )
        return first_6_months + last_6_months



from policyengine_us.model_api import *


class meets_snap_general_work_requirements(Variable):
    value_type = bool
    entity = Person
    label = (
        "Person is eligible for SNAP benefits via general work requirements"
    )
    definition_period = MONTH
    reference = "https://www.law.cornell.edu/cfr/text/7/273.7"

    def formula(person, period, parameters):
        p = parameters(period).gov.usda.snap.work_requirements.general
        age = person("monthly_age", period)
        weekly_hours_worked = person(
            "weekly_hours_worked_before_lsr", period.this_year
        )
        # Under 16 or 60 years of age or older are exempted
        worked_exempted_age = p.age_threshold.exempted.calc(age)
        # Unable to work due to a physical or mental limitation
        is_disabled = person("is_disabled", period)
        # Taking care of a child under six or an incapacitated person
        is_dependent = person("is_tax_unit_dependent", period)
        is_child = age < p.age_threshold.caring_dependent_child
        has_child = person.spm_unit.any(is_dependent & is_child)
        has_incapacitated_person = person.spm_unit.any(
            person("is_incapable_of_self_care", period)
        )
        # Work at least 30 hours a week
        is_working = weekly_hours_worked >= p.weekly_hours_threshold

        return (
            worked_exempted_age
            | is_disabled
            | has_child
            | has_incapacitated_person
            | is_working
        )



from policyengine_us.model_api import *


class is_usda_disabled(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    documentation = "Disabled according to USDA criteria"
    label = "USDA disabled status"

    def formula(person, period, parameters):
        programs = parameters(period).gov.usda.disabled_programs
        return np.any([person(program, period) for program in programs])



from policyengine_us.model_api import *


class IllinoisAABDInstitutionalStatus(Enum):
    A = "A"  # Public Institution (Ineligible)
    B = "B"  # Institution for Mental Diseases (IMD) Ages 22-64 (Ineligible)
    C = "C"  # Public Educational/Vocational Training Institution (MANG Only)
    D = "D"  # Correctional/Penal Institution (Ineligible)
    E = "E"  # Certified Private Psychiatric Hospital (Age 65+)
    F = "F"  # Accredited Private Psychiatric Hospital (Under Age 21)
    G = "G"  # Private Institution (Life Care Contract, Ineligible)
    H = "H"  # Private Institution (Prepaid Care Not Consumed, Ineligible)
    I = "I"  # Non-Compliant Facility (Ineligible)
    NONE = "None"


class il_aabd_institutional_status(Variable):
    value_type = Enum
    entity = SPMUnit
    possible_values = IllinoisAABDInstitutionalStatus
    default_value = IllinoisAABDInstitutionalStatus.NONE
    definition_period = MONTH
    label = "Illinois AABD Institutional Status"
    defined_for = StateCode.MA
    reference = "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-113.70"



from policyengine_us.model_api import *


class ar_deduction_joint(Variable):
    value_type = float
    entity = Person
    label = "Arkansas deduction when married filing jointly"
    unit = USD
    definition_period = YEAR
    reference = "https://www.dfa.arkansas.gov/images/uploads/incomeTaxOffice/2022_AR1000F_and_AR1000NR_Instructions.pdf#page=14"
    defined_for = StateCode.AR

    def formula(person, period, parameters):
        itemized = person("ar_itemized_deductions_joint", period)
        standard = person("ar_standard_deduction_joint", period)
        return max_(itemized, standard)



from policyengine_us.model_api import *


class ia_standard_deduction_indiv(Variable):
    value_type = float
    entity = Person
    label = "Iowa standard deduction when married couples file separately"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.iowa.gov/sites/default/files/2022-01/IA1040%2841-001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2021%20Expanded%20Instructions_010323.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2022IA1040%2841001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-03/2022%20Expanded%20Instructions_022023.pdf"
    )
    defined_for = StateCode.IA

    def formula(person, period, parameters):
        us_filing_status = person.tax_unit("filing_status", period)
        fsvals = us_filing_status.possible_values
        filing_status = select(
            [
                us_filing_status == fsvals.JOINT,
                us_filing_status == fsvals.SINGLE,
                us_filing_status == fsvals.SEPARATE,
                us_filing_status == fsvals.HEAD_OF_HOUSEHOLD,
                us_filing_status == fsvals.SURVIVING_SPOUSE,
            ],
            [
                fsvals.SEPARATE,  # couples are filing separately on Iowa form
                fsvals.SINGLE,
                fsvals.SEPARATE,
                fsvals.HEAD_OF_HOUSEHOLD,
                fsvals.SURVIVING_SPOUSE,
            ],
        )
        p = parameters(period).gov.states.ia.tax.income.deductions.standard

        if p.applies_federal:
            fed_p = parameters(period).gov.irs.deductions
            deduction = fed_p.standard.amount[filing_status]
        else:
            deduction = p.amount[filing_status]

        return person("is_tax_unit_head_or_spouse", period) * deduction



from policyengine_us.model_api import *


class ne_cdcc_refundable_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the Nebraska refundable CDCC"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://revenue.nebraska.gov/files/doc/tax-forms/2021/f_2441n.pdf"
        "https://revenue.nebraska.gov/sites/revenue.nebraska.gov/files/doc/Form_2441N_Ne_Child_and_Dependent_Care_Expenses_8-618-2022_final_2.pdf"
    )
    defined_for = StateCode.NE

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ne.tax.income.credits.cdcc.refundable
        us_agi = tax_unit("adjusted_gross_income", period)
        return us_agi < p.income_limit



from policyengine_us.model_api import *
from policyengine_core.periods import period as period_


def create_family_security_act_2024_eitc() -> Reform:

    class eitc_maximum(Variable):
        value_type = float
        entity = TaxUnit
        label = "Maximum EITC"
        definition_period = YEAR
        reference = "https://www.law.cornell.edu/uscode/text/26/32#a"
        unit = USD

        def formula(tax_unit, period, parameters):
            child_count = tax_unit("eitc_child_count", period)
            p = parameters(
                period
            ).gov.contrib.congress.romney.family_security_act_2_0.eitc.amount
            filing_status = tax_unit("filing_status", period)
            joint = filing_status == filing_status.possible_values.JOINT
            return where(
                joint, p.joint.calc(child_count), p.single.calc(child_count)
            )

    class reform(Reform):
        def apply(self):
            self.update_variable(eitc_maximum)

    return reform


def create_family_security_act_2024_eitc_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_family_security_act_2024_eitc()

    # Look ahead for the next five years
    p = parameters.gov.contrib.congress.romney.family_security_act_2_0.eitc
    reform_active = False
    current_period = period_(period)

    for i in range(5):
        if p(current_period).apply_eitc_structure:
            reform_active = True
            break
        current_period = current_period.offset(1, "year")

    if reform_active:
        return create_family_security_act_2024_eitc()
    else:
        return None


family_security_act_2024_eitc = create_family_security_act_2024_eitc_reform(
    None, None, bypass=True
)



from policyengine_us.model_api import *


class vt_medical_expense_deduction(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "Vermont medical expense deduction"
    reference = (
        "https://legislature.vermont.gov/statutes/section/32/151/05811"  # Titl. 32 V.S.A. § 5811(21)(C)(iv)
        "https://tax.vermont.gov/sites/tax/files/documents/IN-112%20Instr-2022.pdf#page=2",  # Instruction for 2022 SCHEDULE IN-112 - MEDICAL DEDUCTION WORKSHEET
    )
    unit = USD
    defined_for = StateCode.VT
    documentation = "Vermont medical expenses deducted from taxable income."

    def formula(tax_unit, period, parameters):
        # Get federal medical expense deduction (Worksheet line 1a).
        # This points to federal Form 1040, Schedule A, Line 4, which is the deduction itself
        # (not the expenses).
        # The law says "an amount equal to the itemized deduction for medical expenses
        # taken at the federal level by the taxpayer" but does not state whether the filer
        # must have itemized on the federal return to claim it.
        # We assume they do not have to, based on the form.
        federal_medical_expense_deduction = tax_unit(
            "medical_expense_deduction", period
        )
        # Get Vermont standard deduction plus personal exemptions (Worksheet line 2).
        vt_standard_deduction = tax_unit("vt_standard_deduction", period)
        vt_personal_exemptions = tax_unit("vt_personal_exemptions", period)
        # Subtract standard deduction plus personal exemptions from vt allowed medical expense deduction and return (Worksheet line 3).
        return max_(
            0,
            federal_medical_expense_deduction
            - vt_standard_deduction
            - vt_personal_exemptions,
        )



from policyengine_us.model_api import *


class snap_countable_earner(Variable):
    value_type = bool
    entity = Person
    label = "Countable income earner"
    documentation = "Whether this person's earned income is counted for SNAP"
    definition_period = MONTH
    reference = (
        "https://www.law.cornell.edu/uscode/text/7/2014, "
        "https://www.law.cornell.edu/cfr/text/7/273.9#c_3"
    )

    def formula(person, period, parameters):
        # Children in K-12 have income excluded
        is_excluded_child = person("snap_excluded_child_earner", period)

        # Federal Work Study participants have income excluded
        is_work_study_participant = person(
            "is_federal_work_study_participant", period.this_year
        )

        return ~(is_excluded_child | is_work_study_participant)



from policyengine_us.model_api import *


class retirement_distributions(Variable):
    value_type = float
    entity = Person
    label = "Retirement account distributions"
    unit = USD
    definition_period = YEAR
    adds = [
        "taxable_retirement_distributions",
        "tax_exempt_retirement_distributions",
    ]



from policyengine_us.model_api import *
from policyengine_core.periods import period as period_
from policyengine_core.periods import instant


def create_ctc_additional_bracket() -> Reform:
    class ctc_child_individual_maximum(Variable):
        value_type = float
        entity = Person
        label = "CTC maximum amount (child) with additional bracket"
        unit = USD
        definition_period = YEAR
        defined_for = "is_tax_unit_dependent"

        def formula(person, period, parameters):
            age = person("age", period)
            p = parameters(period).gov.contrib.ctc.additional_bracket.amount
            return p.base.calc(age)

    class ctc_refundable_maximum(Variable):
        value_type = float
        entity = TaxUnit
        label = "Maximum refundable CTC"
        unit = USD
        documentation = "The maximum refundable CTC for this person."
        definition_period = YEAR

        def formula(tax_unit, period, parameters):
            person = tax_unit.members
            child_amount = max_(
                person("ctc_child_individual_maximum", period),
                person("ctc_child_individual_maximum_arpa", period),
            )
            # Use either normal or ARPA CTC maximums.
            age = person("age", period)
            is_dependent = person("is_tax_unit_dependent", period)
            p = parameters(period).gov.contrib.ctc.additional_bracket.amount
            refundable_max = p.actc.calc(age) * is_dependent
            return tax_unit.sum(min_(child_amount, refundable_max))

    class reform(Reform):
        def apply(self):
            self.update_variable(ctc_child_individual_maximum)
            self.update_variable(ctc_refundable_maximum)

    return reform


def create_ctc_additional_bracket_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_ctc_additional_bracket()

    p = parameters.gov.contrib.ctc.additional_bracket

    reform_active = False
    current_period = period_(period)

    for i in range(5):
        if p(current_period).in_effect:
            reform_active = True
            break
        current_period = current_period.offset(1, "year")

    if reform_active:
        return create_ctc_additional_bracket()
    else:
        return None


ctc_additional_bracket = create_ctc_additional_bracket_reform(
    None, None, bypass=True
)



from policyengine_us.model_api import *


class mn_elderly_disabled_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Minnesota elderly/disabled subtraction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1r_21.pdf"
        "https://www.revenue.state.mn.us/sites/default/files/2023-01/m1r_22.pdf"
    )
    defined_for = StateCode.MN

    def formula(tax_unit, period, parameters):
        mn_itax = parameters(period).gov.states.mn.tax.income
        p = mn_itax.subtractions.elderly_disabled
        filing_status = tax_unit("filing_status", period)
        # calculate the subtraction amount
        # ... determine age eligibility
        person = tax_unit.members
        is_elderly = person("age", period) >= p.minimum_age
        is_head = person("is_tax_unit_head", period)
        elderly_head = is_head & is_elderly
        is_spouse = person("is_tax_unit_spouse", period)
        elderly_spouse = is_spouse & is_elderly
        # ... determine disability eligiblity
        is_disabled = person("is_permanently_and_totally_disabled", period)
        disabled_head = is_head & is_disabled
        disabled_spouse = is_spouse & is_disabled
        # ... determine base subtraction amount
        base_amount = where(
            tax_unit.any(elderly_head | elderly_spouse),
            p.base_amount[filing_status],
            0,  # because federally-taxable disability income
            # included in federal Form 1040, line 1z, is unknown
        )
        # ... determine disability income as on federal Schedule R
        is_dependent = person("is_tax_unit_dependent", period)
        dinc_person = person("total_disability_payments", period)
        disinc_person = dinc_person * ~is_dependent * ~is_elderly
        disability_income = tax_unit.sum(disinc_person)
        # ... determine untaxed social security benefits
        unit_socsec = add(tax_unit, period, ["social_security"])
        unit_taxed_socsec = add(tax_unit, period, ["taxable_social_security"])
        untaxed_social_security = unit_socsec - unit_taxed_socsec
        # ... determine total disability and social security benefits
        benefits = disability_income + untaxed_social_security
        # ... subtract benefits from base amount
        amount = max_(0, base_amount - benefits)
        # ... determine net agi
        fed_agi = tax_unit("adjusted_gross_income", period)
        agi = fed_agi + tax_unit("form_4972_lumpsum_distributions", period)
        agi_offset_base = p.agi_offset_base[filing_status]
        joint = filing_status == filing_status.possible_values.JOINT
        head_eligible = tax_unit.any(elderly_head | disabled_head)
        spouse_eligible = tax_unit.any(elderly_spouse | disabled_spouse)
        eligible = head_eligible | spouse_eligible
        joint_with_two_eligibles = joint & head_eligible & spouse_eligible
        agi_offset_extra = joint_with_two_eligibles * p.agi_offset_extra
        agi_offset = agi_offset_base + agi_offset_extra
        net_agi = max_(0, agi - agi_offset)
        fraction_of_net_agi = p.net_agi_fraction * net_agi
        # ... subtract fracton_of_net_agi from amount to get final amount
        final_amount = max_(0, amount - fraction_of_net_agi)
        return eligible * final_amount



from policyengine_us.model_api import *


class tx_ottanf(Variable):
    value_type = float
    entity = SPMUnit
    label = "Texas One-Time TANF (OTTANF) payment"
    unit = USD
    definition_period = MONTH
    reference = "https://www.hhs.texas.gov/handbooks/texas-works-handbook/a-2410-general-policy"
    defined_for = "tx_ottanf_eligible"

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.tx.tanf.ottanf

        # One-time payment of $1,000 divided by 12 for monthly equivalent
        return p.payment_amount / MONTHS_IN_YEAR



from policyengine_us.model_api import *


class or_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "OR taxable income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.OR

    def formula(tax_unit, period, parameters):
        or_agi = tax_unit("or_agi", period)
        deductions = tax_unit("or_deductions", period)
        return max_(or_agi - deductions, 0)



from policyengine_us.model_api import *


class is_ebb_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for Emergency Broadband Benefit"
    documentation = "Eligible for Emergency Broadband Benefit"
    definition_period = YEAR

    def formula(spm_unit, period, parameters):
        programs = parameters(period).gov.fcc.ebb.categorical_eligibility
        eligible = np.any([spm_unit(program, period) for program in programs])
        # In transition period to Affordable Connectivity Program, households
        # must already be enrolled to receive EBB.
        if parameters(period).gov.fcc.ebb.prior_enrollment_required:
            return eligible & spm_unit("enrolled_in_ebb", period)
        return eligible



from policyengine_us.model_api import *


class wa_tanf_payment_standard(Variable):
    value_type = float
    entity = SPMUnit
    label = "Washington TANF payment standard"
    unit = USD
    definition_period = MONTH
    reference = "https://app.leg.wa.gov/wac/default.aspx?cite=388-478-0020"
    defined_for = StateCode.WA

    def formula(spm_unit, period, parameters):
        # Get payment standard by family size per WAC 388-478-0020
        p = parameters(period).gov.states.wa.dshs.tanf

        # Get assistance unit size, capped at maximum table value (10)
        size = spm_unit("spm_unit_size", period)
        size_capped = min_(size, p.maximum_family_size)

        # Look up payment standard for this family size
        return p.payment_standard.amount[size_capped]



from policyengine_us.model_api import *


class dc_deduction_indiv(Variable):
    value_type = float
    entity = Person
    label = "DC deduction for each person in tax unit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/52926_D-40_12.21.21_Final_Rev011122.pdf#page=44"
        "https://otr.cfo.dc.gov/sites/default/files/dc/sites/otr/publication/attachments/2022_D-40_Booklet_Final_blk_01_23_23_Ordc.pdf#page=44"
    )
    defined_for = StateCode.DC

    def formula(person, period, parameters):
        tax_unit_deduction = person.tax_unit("dc_deduction_joint", period)
        # The above references say the following:
        # "You may allocate this [tax-unit deduction] amount as you wish."
        # Here we allocate in proportion to head and spouse DC AGI
        person_agi = person("dc_agi", period)
        tax_unit_agi = person.tax_unit.sum(person_agi)
        share = np.zeros_like(tax_unit_agi)
        mask = tax_unit_agi > 0
        share[mask] = person_agi[mask] / tax_unit_agi[mask]
        return share * tax_unit_deduction



from policyengine_us.model_api import *


class ga_taxable_income(Variable):
    value_type = float
    entity = TaxUnit
    label = "Georgia taxable income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://dor.georgia.gov/it-511-individual-income-tax-booklet"
        # above reference provides access to booklets for all years
        # definition of Georgia taxable income starts on page 12
    )
    defined_for = StateCode.GA

    def formula(tax_unit, period, parameters):
        agi = tax_unit("ga_agi", period)
        deductions = tax_unit("ga_deductions", period)
        exemptions = tax_unit("ga_exemptions", period)
        return max_(0, agi - deductions - exemptions)



from policyengine_us.model_api import *


class receives_housing_assistance(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Currently receives housing assistance"
    documentation = "Currently receives housing assistance"
    definition_period = YEAR



from policyengine_us.model_api import *


class base_part_b_premium(Variable):
    value_type = float
    entity = Person
    label = "Medicare Part B Base Premium"
    unit = USD
    documentation = "Medicare Part B Premium."
    definition_period = YEAR
    defined_for = "is_medicare_eligible"

    def formula(person, period, parameters):
        p = parameters(period).gov.hhs.medicare.part_b
        # base_premium is a monthly amount, multiply by MONTHS_IN_YEAR for annual
        return p.base_premium * MONTHS_IN_YEAR



from policyengine_us.model_api import *


class current_home_energy_use(Variable):
    value_type = float
    entity = Household
    label = "Current home energy use in monthly kilowatt hours"
    unit = "kWh/month"
    definition_period = YEAR
    reference = "https://www.democrats.senate.gov/imo/media/doc/inflation_reduction_act_of_2022.pdf#page=587"



from policyengine_us.model_api import *


class il_hbwd_asset_eligible(Variable):
    value_type = bool
    entity = Person
    label = (
        "Illinois Health Benefits for Workers with Disabilities asset eligible"
    )
    definition_period = MONTH
    reference = (
        "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-120.510",
        "https://hfs.illinois.gov/medicalprograms/hbwd/eligibility.html",
    )
    defined_for = StateCode.IL

    def formula(person, period, parameters):
        p = parameters(period).gov.states.il.hfs.hbwd.eligibility
        # Check countable assets against $25,000 limit
        countable_assets = person("il_hbwd_countable_assets", period)
        return countable_assets <= p.asset.limit



from policyengine_us.model_api import *


class ca_ala_general_assistance_base_amount(Variable):
    value_type = float
    entity = SPMUnit
    unit = USD
    label = "Alameda County General Assistance base amount"
    definition_period = MONTH
    defined_for = "ca_ala_general_assistance_eligible_person"
    reference = "https://www.alamedacountysocialservices.org/acssa-assets/PDF/GA-Policies/GA-Regulations.pdf#page=28"

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.local.ca.ala.general_assistance.amount
        eligible_persons = spm_unit.members(
            "ca_ala_general_assistance_eligible_person", period
        )
        num_eligible = spm_unit.sum(eligible_persons)
        return where(num_eligible == 2, p.married, p.single)



from policyengine_us.model_api import *


class co_care_worker_credit_eligible_care_worker_count(Variable):
    value_type = float
    entity = TaxUnit
    label = "Number of eligible Care Workers for the Colorado Care Worker Tax Credit"
    defined_for = StateCode.CO
    definition_period = YEAR
    reference = "https://law.justia.com/codes/colorado/title-39/specific-taxes/income-tax/article-22/part-5/section-39-22-566/"

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        care_worker = person(
            "co_care_worker_credit_eligible_care_worker", period
        )
        head_or_spouse = person("is_tax_unit_head_or_spouse", period)
        return tax_unit.sum(care_worker & head_or_spouse)



from policyengine_us.model_api import *


class la_surviving_spouse_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "Louisiana qualifying surviving spouse exemption"
    reference = (
        "https://www.revenue.louisiana.gov/taxforms/6935(11_02)F.pdf#page=1"
    )
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.LA

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        is_surviving_spouse = tax_unit.any(
            person("is_surviving_spouse", period)
        )
        p = parameters(period).gov.states.la.tax.income.exemptions
        return is_surviving_spouse * p.surviving_spouse



from policyengine_us.model_api import *


class id_household_and_dependent_care_expense_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Idaho household and dependent care expense deduction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.ID

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.id.tax.income.deductions.dependent_care_expenses
        expenses = tax_unit("tax_unit_childcare_expenses", period)
        # In 2023 Idaho implemented an increased cap independent of the number of children
        limit = max_(tax_unit("id_cdcc_limit", period), p.cap)
        eligible_capped_expenses = min_(expenses, limit)
        # cap further to the lowest earnings between the taxpayer and spouse
        return min_(
            eligible_capped_expenses,
            tax_unit("min_head_spouse_earned", period),
        )



from policyengine_us.model_api import *


class wi_capital_gain_loss_addition(Variable):
    value_type = float
    entity = TaxUnit
    label = "WI capital gain/loss addition to federal adjusted gross income"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://www.revenue.wi.gov/TaxForms2021/2021-ScheduleAD-inst.pdf#page=2"
        "https://www.revenue.wi.gov/TaxForms2022/2022-ScheduleAD-Inst.pdf#page=2"
    )
    defined_for = StateCode.WI

    def formula(tax_unit, period, parameters):
        us_capital_loss = tax_unit("limited_capital_loss", period)
        wi_capital_loss = tax_unit("wi_capital_loss", period)
        return max_(0, us_capital_loss - wi_capital_loss)



from policyengine_us.model_api import *


class snap_utility_allowance(Variable):
    value_type = float
    entity = SPMUnit
    label = "Standard Utility Allowance"
    unit = USD
    documentation = "The regular utility allowance deduction for SNAP"
    definition_period = MONTH

    adds = [
        "snap_standard_utility_allowance",
        "snap_limited_utility_allowance",
        "snap_individual_utility_allowance",
    ]



from policyengine_us.model_api import *


class il_liheap(Variable):
    value_type = float
    entity = SPMUnit
    label = "Illinois LIHEAP total benefit"
    unit = USD
    definition_period = YEAR
    defined_for = "il_liheap_eligible"
    reference = "https://dceo.illinois.gov/communityservices/utilitybillassistance.html"

    adds = ["il_liheap_base_payment", "il_liheap_crisis_assistance_amount"]



from policyengine_us.model_api import *


class takes_up_aca_if_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Whether a random eligible SPM unit does not claim ACA Premium Tax Credit"
    definition_period = YEAR

    def formula(tax_unit, period, parameters):
        seed = tax_unit("aca_take_up_seed", period)
        takeup_rate = parameters(period).gov.aca.takeup_rate
        return seed < takeup_rate



from policyengine_us.model_api import *


class tanf_gross_unearned_income(Variable):
    value_type = float
    entity = Person
    label = (
        "Temporary Assistance for Needy Families (TANF) gross unearned income"
    )
    unit = USD
    definition_period = MONTH

    adds = "gov.hhs.tanf.cash.income.sources.unearned"



from policyengine_us.model_api import *


class ny_cdcc(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY CDCC"
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/606"  # (c)
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        cdcc_max = tax_unit("ny_cdcc_max", period)
        expenses = tax_unit("cdcc_relevant_expenses", period)
        cdcc_rate = tax_unit("ny_cdcc_rate", period) * tax_unit(
            "cdcc_rate", period
        )
        return min_(cdcc_max, expenses * cdcc_rate)



from policyengine_us.model_api import *


class ga_standard_deduction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Georgia standard deduction"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://apps.dor.ga.gov/FillableForms/PDFViewer/Index?form=2022GA500"
        "https://advance.lexis.com/documentpage/?pdmfid=1000516&crid=fb5db531-a80f-4790-bddb-eefc8327ef60&config=00JAA1MDBlYzczZi1lYjFlLTQxMTgtYWE3OS02YTgyOGM2NWJlMDYKAFBvZENhdGFsb2feed0oM9qoQOMCSJFX5qkd&pddocfullpath=%2Fshared%2Fdocument%2Fstatutes-legislation%2Furn%3AcontentItem%3A65D2-CDH3-CGX8-044N-00008-00&pdcontentcomponentid=234186&pdteaserkey=sr1&pditab=allpods&ecomp=8s65kkk&earg=sr1&prid=66f02b0a-c5ae-4162-9535-127751546807"
    )
    defined_for = StateCode.GA

    def formula(tax_unit, period, parameters):
        # person = tax_unit.members
        p = parameters(period).gov.states.ga.tax.income.deductions.standard
        filing_status = tax_unit("filing_status", period)
        base = p.amount[filing_status]
        # total extra deduction
        if p.applies:
            additional_standard = tax_unit(
                "ga_additional_standard_deduction", period
            )
            return base + additional_standard
        return base



from policyengine_us.model_api import *


class nj_total_income(Variable):
    value_type = float
    entity = Person
    label = "New Jersey total income"
    unit = USD
    documentation = "New Jersey total income calculated as gross income plus additions minus subtractions per NJ Statute 54A:5-1. This is the income base before exclusions."
    definition_period = YEAR
    reference = (
        "https://law.justia.com/codes/new-jersey/title-54a/section-54a-5-1/",
        "https://www.nj.gov/treasury/taxation/pdf/current/1040.pdf",
    )
    defined_for = StateCode.NJ

    def formula(person, period, parameters):
        # NJ Total Income = Gross Income + Additions - Subtractions
        # Per NJ Statute 54A:5-1 and Form NJ-1040
        gross_income = person("nj_gross_income", period)
        additions = person("nj_additions", period)
        subtractions = person("nj_agi_subtractions", period)
        return max_(0, gross_income + additions - subtractions)



from policyengine_us.model_api import *


class detailed_occupation_recode(Variable):
    value_type = int
    entity = Person
    label = "CPS detailed occupation recode of previous year"
    documentation = "This variable is the POCCU2 variable in the Current Population Survey."
    definition_period = YEAR



from policyengine_us.model_api import *


class receives_medicaid(Variable):
    value_type = bool
    entity = Person
    label = "Currently receives Medicaid"
    definition_period = YEAR
    documentation = """
    Input variable indicating a person currently has Medicaid coverage.
    Used to model continuous coverage for existing enrollees who may no
    longer meet eligibility criteria under new rules (e.g., CA undocumented
    immigrant enrollment freeze starting 2026).

    This differs from medicaid_enrolled which is computed based on
    eligibility and takeup rate.
    """



from policyengine_us.model_api import *


class pell_grant(Variable):
    value_type = float
    entity = Person
    label = "Pell Grant"
    definition_period = YEAR
    unit = USD

    def formula(person, period, parameters):
        coa = person("cost_of_attending_college", period)
        months_in_school = person("pell_grant_months_in_school", period)
        efc = person("pell_grant_efc", period)
        sai = person("pell_grant_sai", period)
        eligibility = person("pell_grant_eligibility_type", period)
        uses_efc = person("pell_grant_uses_efc", period)
        uses_sai = person("pell_grant_uses_sai", period)
        p = parameters(period).gov.ed.pell_grant
        contribution = select([uses_efc, uses_sai], [efc, sai])
        unbounded = coa - contribution
        amount = where(unbounded < p.amount.min, 0, unbounded)
        uncapped_efc_pell = amount * (
            months_in_school / p.months_in_school_year
        )
        uncapped_sai_pell = select(
            [
                eligibility == eligibility.possible_values.INELIGIBLE,
                eligibility == eligibility.possible_values.MAXIMUM,
                eligibility == eligibility.possible_values.MINIMUM,
            ],
            [0, p.amount.max, amount],
        )
        uncapped = select(
            [uses_efc, uses_sai], [uncapped_efc_pell, uncapped_sai_pell]
        )
        max = min_(coa, p.amount.max)
        return min_(max, uncapped)



from policyengine_us.model_api import *


class nyc_school_tax_credit_rate_reduction_amount_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for NYC School Tax Credit Rate Reduction Amount"
    definition_period = YEAR
    defined_for = "in_nyc"

    def formula(tax_unit, period, parameters):
        # First get their NYC taxable income.
        nyc_taxable_income = tax_unit("nyc_taxable_income", period)

        # Then get the School Tax Credit rate reduction amount part of the parameter tree.
        p = parameters(
            period
        ).gov.local.ny.nyc.tax.income.credits.school.rate_reduction

        return nyc_taxable_income <= p.income_limit



from policyengine_us.model_api import *


class md_local_income_tax_before_credits(Variable):
    value_type = float
    entity = TaxUnit
    label = "MD local income tax before credits"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        county = tax_unit.household("county_str", period)
        in_md = tax_unit.household("state_code_str", period) == "MD"
        rate = np.zeros_like(county, dtype=float)
        rates = parameters(period).gov.local.tax.income.flat_tax_rate
        if in_md.sum() > 0:
            # Only compute Maryland local taxes if the simulation includes Maryland households.
            rate[in_md] = rates[county[in_md]]
        return rate * tax_unit("md_taxable_income", period)



from policyengine_us.model_api import *


class az_increased_standard_deduction_for_charitable_contributions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arizona increased standard deduction for charitable contributions"
    unit = USD
    documentation = "https://www.azleg.gov/viewdocument/?docName=https://www.azleg.gov/ars/43/01041.htm"
    definition_period = YEAR
    defined_for = StateCode.AZ

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.az.tax.income.deductions.standard.increased
        charitable_deduction = tax_unit("charitable_deduction", period)
        charitable_contributions_credit = tax_unit(
            "az_charitable_contributions_credit", period
        )
        charitable_deduction_after_credit = max_(
            charitable_deduction - charitable_contributions_credit, 0
        )
        return p.rate * charitable_deduction_after_credit



from policyengine_us.model_api import *


class md_aged_dependent_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "MD aged dependent exemption"
    unit = USD
    definition_period = YEAR
    reference = "https://govt.westlaw.com/mdc/Document/NF59A76006EA511E8ABBEE50DE853DFF4?viewType=FullText&originationContext=documenttoc&transitionType=CategoryPageItem&contextData=(sc.Default)"
    defined_for = StateCode.MD

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.md.tax.income.exemptions.aged
        # These apply to dependents over the age of 65
        person = tax_unit.members
        dependent = person("is_tax_unit_dependent", period)
        age = person("age", period)
        elderly = age >= p.age
        aged_dependents = tax_unit.sum(dependent & elderly)
        return aged_dependents * p.aged_dependent



from policyengine_us.model_api import *


class count_distinct_utility_expenses(Variable):
    value_type = int
    entity = SPMUnit
    label = "Number of distinct utility expenses"
    documentation = "The number of distinct utility expenses."
    definition_period = YEAR

    def formula(spm_unit, period, parameters):
        UTILITIES = [
            "heating_cooling",
            # Use pre-subsidy expenses to avoid circular references
            # since electricity subsidies depend on SNAP enrollment.
            "pre_subsidy_electricity",
            "gas",
            "phone",
            "trash",
            "water",
            "sewage",
        ]
        return sum(
            [
                spm_unit(variable + "_expense", period) > 0
                for variable in UTILITIES
            ]
        )



from policyengine_us.model_api import *


class per_capita_chip(Variable):
    value_type = float
    entity = Person
    label = "Average CHIP payment"
    unit = USD
    documentation = "Per-capita CHIP payment for this person's State."
    definition_period = YEAR
    reference = "https://www.macpac.gov/publication/chip-spending-by-state/"
    defined_for = "is_chip_eligible"

    def formula(person, period, parameters):
        state_code = person.household("state_code", period)
        p = parameters(period).calibration.gov.hhs.cms.chip

        spending = p.spending.total.total[state_code]
        enrollment = p.enrollment.total[state_code]

        per_capita = np.zeros_like(enrollment, dtype=float)
        mask = enrollment > 0
        per_capita[mask] = spending[mask] / enrollment[mask]
        return per_capita



from policyengine_us.model_api import *


class nyc_school_tax_credit(Variable):
    value_type = float
    entity = TaxUnit
    label = "NYC School Tax Credit"
    unit = USD
    definition_period = YEAR
    defined_for = "in_nyc"

    adds = [
        "nyc_school_tax_credit_fixed_amount",
        "nyc_school_tax_credit_rate_reduction_amount",
    ]



from policyengine_us.model_api import *


class is_cdcc_eligible(Variable):
    value_type = bool
    entity = Person
    label = "CDCC-eligible"
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/21#b_1"

    def formula(person, period, parameters):
        age = person("age", period)
        # Subsection A.
        max_age = parameters(period).gov.irs.credits.cdcc.eligibility.child_age
        qualifies_by_age = age < max_age
        # Subsection B (dependent) and C (spouse).
        non_head = ~person("is_tax_unit_head", period)
        disabled = person("is_incapable_of_self_care", period)
        qualifies_by_disability = non_head & disabled
        return qualifies_by_age | qualifies_by_disability



from policyengine_us.model_api import *


class il_aabd_asset_value_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    definition_period = MONTH
    label = "Eligible for Illinois Aid to the Aged, Blind or Disabled (AABD) due to asset"
    reference = (
        "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-113.142",
    )
    defined_for = StateCode.IL

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.il.dhs.aabd.asset.disregard
        size = spm_unit("spm_unit_size", period)
        countable_asset = spm_unit("il_aabd_countable_assets", period)
        capped_size = min_(size, 2)
        reduced_unit_size = max_(size - 2, 0)
        p1 = p.base[capped_size]
        pn = p.additional
        asset_disregard = p1 + pn * reduced_unit_size

        return countable_asset <= asset_disregard



from policyengine_us.model_api import *


class is_tanf_enrolled(Variable):
    value_type = bool
    entity = SPMUnit
    definition_period = MONTH
    label = "Enrolled in TANF"
    documentation = "Whether the familiy is currently enrolled in the Temporary Assistance for Needy Families program."



from policyengine_us.model_api import *


class ky_income_tax_before_non_refundable_credits_indiv(Variable):
    value_type = float
    entity = Person
    label = "Kentucky income tax before non-refundable credits when married couples are filing separately"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.KY

    def formula(person, period, parameters):
        ky_taxable_income_indiv = person("ky_taxable_income_indiv", period)
        p = parameters(period).gov.states.ky.tax.income
        return ky_taxable_income_indiv * p.rate



from policyengine_us.model_api import *


def create_in_state_variable(state: str) -> Type[Variable]:
    return type(
        state,
        (Variable,),
        {
            "definition_period": YEAR,
            "label": f"In {state}",
            "value_type": bool,
            "entity": Household,
            "formula": in_state(state),
            "module_name": "",
            "hidden_input": True,
        },
    )


def create_50_state_variables() -> List[Type[Variable]]:
    return [create_in_state_variable(state) for state in STATES]



from policyengine_us.model_api import *


class ca_eitc_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "CalEITC eligible"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.CA

    def formula(tax_unit, period, parameters):
        person = tax_unit.members
        p = parameters(period).gov.states.ca.tax.income.credits.earned_income

        age = person("age", period)
        is_dependent = person("is_tax_unit_dependent", period)

        meets_age_requirements = tax_unit.any(
            (age >= p.eligibility.age.min)
            & (age <= p.eligibility.age.max)
            & ~is_dependent
        )

        eitc_investment_income = tax_unit(
            "eitc_relevant_investment_income", period
        )

        meets_investment_income_requirements = (
            eitc_investment_income <= p.eligibility.max_investment_income
        )

        return meets_age_requirements & meets_investment_income_requirements



from policyengine_us.model_api import *


class ca_ffyp_eligible(Variable):
    value_type = bool
    entity = Person
    label = "Eligible person for the California Former Foster Youth Program"
    definition_period = YEAR
    reference = (
        "https://www.dhcs.ca.gov/services/medi-cal/eligibility/Pages/FFY_Bene.aspx",
        "https://www.dhcs.ca.gov/formsandpubs/forms/Forms/MCED/MC_Forms/MC250A_Eng.pdf",
    )
    defined_for = StateCode.CA

    def formula(person, period, parameters):
        p = parameters(period).gov.states.ca.dhcs.ffyp
        age = person("age", period)
        age_eligible = (age > p.foster_care_age_minimum) & (
            age < p.age_threshold
        )  # Person must be below age limit and previously in foster care for a valid age
        was_in_foster_care = person(
            "was_in_foster_care", period
        )  # Assumes the person was in foster care for a valid age
        return age_eligible & was_in_foster_care



from policyengine_us.model_api import *


class or_wfhdc_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the Oregon working family household and dependent care credit"
    documentation = "Oregon Working Family Household and Dependent Care Credit household eligibility"
    definition_period = YEAR
    reference = (
        "https://www.oregon.gov/dor/forms/FormsPubs/schedule-or-wfhdc-inst_101-195-1_2022.pdf#pahe=1",
        "https://law.justia.com/codes/oregon/2021/volume-08/chapter-315/section-315-264/",
    )
    defined_for = StateCode.OR

    def formula(tax_unit, period, parameters):
        household_size_eligible = tax_unit(
            "or_wfhdc_household_size_eligible", period
        )
        income_eligible = tax_unit("or_wfhdc_income_eligible", period)
        has_qualified_individual = tax_unit(
            "or_wfhdc_has_qualified_individual_eligible", period
        )
        employment_eligible = tax_unit("or_wfhdc_employment_eligible", period)

        return (
            household_size_eligible
            & income_eligible
            & has_qualified_individual
            & employment_eligible
        )



from policyengine_us.model_api import *


class commodity_supplemental_food_program_eligible(Variable):
    value_type = bool
    entity = Person
    definition_period = YEAR
    label = "Commodity Supplemental Food Program eligible"

    def formula(person, period, parameters):
        p = parameters(period).gov.usda.csfp
        # CFR defines CSFP income similarly to WIC income.
        # Assume resources are counted at the SPM unit level.
        fpg = person.spm_unit("school_meal_fpg_ratio", period)
        age = person("age", period)

        age_eligible = age >= p.min_age
        federal_income_eligible = fpg <= p.fpg_limit
        tx_income_eligible = person("tx_dta_csfp_income_eligible", period)
        state_code = person.household("state_code", period)
        in_tx = state_code == StateCode.TX
        income_eligible = where(
            in_tx, tx_income_eligible, federal_income_eligible
        )
        return age_eligible & income_eligible



"""
This makes the function create_tax_employer_payroll_reform available for
use without needing to specify the full module path.
i.e., we can just call:
from policyengine_us.reforms.tax_employer_payroll import create_tax_employer_payroll_reform
instead of:
from policyengine_us.reforms.tax_employer_payroll.tax_employer_payroll import create_tax_employer_payroll_reform
"""

from .tax_employer_social_security_tax import (
    create_tax_employer_social_security_tax_reform,
)

from .tax_employer_medicare_tax import (
    create_tax_employer_medicare_tax_reform,
)

from .tax_employer_payroll_tax import (
    create_tax_employer_payroll_tax_reform,
)
from .non_refundable_ss_credit import (
    create_non_refundable_ss_credit_reform,
)
from .senior_deduction_extension import (
    create_senior_deduction_extension_reform,
)



from policyengine_us.model_api import *


class ma_part_b_taxable_income_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "MA Part B taxable income exemption"
    unit = USD
    definition_period = YEAR
    reference = "https://www.mass.gov/service-details/view-massachusetts-personal-income-tax-exemption"
    defined_for = StateCode.MA

    def formula(tax_unit, period, parameters):
        tax = parameters(period).gov.states.ma.tax.income
        # (B)(a)(3): Taxes for retirement programs.
        # NB: The law only mentions FICA and FRRA, but mass.gov includes SECA.
        # https://www.mass.gov/service-details/learn-about-business-and-professional-income
        person = tax_unit.members
        filing_status = tax_unit("filing_status", period)
        # (B)(b): Exemptions.
        # (1A) and (2A): Personal exemption based on filing status.
        personal_exemption = tax.exemptions.personal[filing_status]
        # (1B) and (2B): Blind exemptions.
        blind = person("is_blind", period)
        dependent = person("is_tax_unit_dependent", period)
        count_blind = tax_unit.sum(~dependent & blind)
        blind_exemption = tax.exemptions.blind * count_blind
        # (1C) and (2C): Aged exemptions.
        age = person("age", period)
        count_aged = tax_unit.sum(
            ~dependent & (age >= tax.exemptions.aged.age)
        )
        aged_exemption = tax.exemptions.aged.amount * count_aged
        # (3): Dependent exemptions.
        count_dependents = tax_unit("tax_unit_dependents", period)
        dependent_exemption = tax.exemptions.dependent * count_dependents
        # (4): Medical expense deduction for itemizers.
        itemizes = tax_unit("tax_unit_itemizes", period)
        federal_medical_expense_deduction = tax_unit(
            "medical_expense_deduction", period
        )
        medical_dental_exemption = itemizes * federal_medical_expense_deduction
        return (
            personal_exemption
            + dependent_exemption
            + aged_exemption
            + blind_exemption
            + medical_dental_exemption
        )



from policyengine_us.model_api import *


class wv_withheld_income_tax(Variable):
    value_type = float
    entity = Person
    label = "West Virginia withheld income tax"
    defined_for = StateCode.WV
    unit = USD
    definition_period = YEAR

    def formula(person, period, parameters):
        agi = person("adjusted_gross_income_person", period)
        p = parameters(period).gov.states.wv.tax.income
        # Since West Virginia does not have a standard deduction, we apply the maximum
        # personal exemption amount
        personal_exmptions = p.exemptions.base_personal + p.exemptions.personal
        reduced_agi = max_(agi - personal_exmptions, 0)
        return p.rates.single.calc(reduced_agi)



from policyengine_us.model_api import *


class ky_cdcc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Kentucky household and dependent care service credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://apps.legislature.ky.gov/law/statutes/statute.aspx?id=29058"
    )
    defined_for = StateCode.KY

    def formula(tax_unit, period, parameters):
        # Kentucky matches the federal credit taken
        dependent_care_credit = tax_unit("cdcc", period)
        rate = parameters(
            period
        ).gov.states.ky.tax.income.credits.dependent_care_service.match
        return dependent_care_credit * rate



from policyengine_us.model_api import *


class pell_grant_simplified_formula_applies(Variable):
    value_type = bool
    entity = Person
    label = "Use Pell Grant simplified formula"
    definition_period = YEAR

    def formula(person, period, parameters):
        tax_unit = person.tax_unit
        head_income = tax_unit("pell_grant_primary_income", period)
        p = parameters(period).gov.ed.pell_grant.efc.simplified
        income_eligible = head_income < p.income_limit
        has_benefits = add(tax_unit, period, p.benefits) > 0
        return income_eligible & has_benefits & p.applies



from policyengine_us.model_api import *


class oh_lump_sum_distribution_credit_person(Variable):
    value_type = float
    entity = Person
    label = "Ohio lump sum distribution credit"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.ohio.gov/static/forms/ohio_individual/individual/2021/pit-it1040-booklet.pdf#page=29",
    )
    defined_for = "oh_lump_sum_distribution_credit_eligible_person"

    def formula(person, period, parameters):
        p = parameters(
            period
        ).gov.states.oh.tax.income.credits.lump_sum_distribution

        age = person("age", period)
        return p.base_amount * p.expected_remaining_years.calc(age)



from policyengine_us.model_api import *


class wv_cdcc(Variable):
    value_type = float
    entity = TaxUnit
    label = "West Virginia Child and Dependent Care Credit"
    unit = USD
    defined_for = StateCode.WV
    definition_period = YEAR
    reference = "https://code.wvlegislature.gov/11-21-26/"

    def formula(tax_unit, period, parameters):
        # West Virginia matched the federal credit taken
        cdcc = tax_unit("cdcc", period)
        p = parameters(period).gov.states.wv.tax.income.credits.cdcc
        return cdcc * p.match



from policyengine_us.model_api import *


class co_care_worker_credit_eligible(Variable):
    value_type = float
    entity = TaxUnit
    label = "Colorado Care Worker Tax Credit eligible"
    defined_for = StateCode.CO
    unit = USD
    definition_period = YEAR
    reference = "https://law.justia.com/codes/colorado/title-39/specific-taxes/income-tax/article-22/part-5/section-39-22-566/"

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.co.tax.income.credits.care_worker
        joint = tax_unit("tax_unit_is_joint", period)
        agi = tax_unit("adjusted_gross_income", period)
        income_limit = where(
            joint, p.income_limit.joint, p.income_limit.non_joint
        )
        income_eligible = agi <= income_limit
        eligible_care_worker_present = (
            tax_unit(
                "co_care_worker_credit_eligible_care_worker_count", period
            )
            > 0
        )
        return income_eligible & eligible_care_worker_present



from policyengine_us.model_api import *


class health_savings_account_ald(Variable):
    value_type = float
    entity = TaxUnit
    label = "Health savings account ALD"
    unit = USD
    documentation = "Above-the-line deduction from gross income for health savings account expenses."
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/223"



from policyengine_us.model_api import *


class il_aabd_aged_blind_disabled_person(Variable):
    value_type = bool
    entity = Person
    definition_period = MONTH
    label = "Aged, blind, or disabled person for Illinois Aid to the Aged, Blind or Disabled (AABD)"
    reference = (
        "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-113.30",
        "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-113.40",
        "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-113.50",
    )
    defined_for = StateCode.IL

    def formula(person, period, parameters):
        p = parameters(period).gov.states.il.dhs.aabd
        age = person("monthly_age", period)
        is_aged = age >= p.aged_age_threshold
        is_blind = person("is_blind", period)
        is_disabled = person("is_ssi_disabled", period)
        return is_aged | is_blind | is_disabled



from policyengine_us.model_api import *


class nyc_cdcc(Variable):
    value_type = float
    entity = TaxUnit
    label = "NYC Child and Dependent Care Credit"
    unit = USD
    definition_period = YEAR
    reference = "https://www.tax.ny.gov/pdf/current_forms/it/it216i.pdf"
    defined_for = "nyc_cdcc_eligible"

    def formula(tax_unit, period, parameters):
        # The NYC CDCC is a share of the NY State CDCC.
        # The share magnitude is determined by the fraction of CDCC
        # relevant expenses used for children under 4 and the NYC CDCC rate,
        # which depends on income.

        # Get their NY State CDCC (line 14 on Form IT-216).
        nys_cdcc = tax_unit("ny_cdcc", period)

        # Get the share of childcare expenses that count towards the NYC CDCC.
        share_of_childcare_expenses_for_children_under_four = tax_unit(
            "nyc_cdcc_share_qualifying_childcare_expenses", period
        )

        # Take this share of the NY State CDCC.
        nyc_qualifying_cdcc_amount = (
            nys_cdcc * share_of_childcare_expenses_for_children_under_four
        )

        # Get the CDCC rate "applicable percentage" portion of the NYS CDCC
        applicable_percentage = tax_unit(
            "nyc_cdcc_applicable_percentage", period
        )

        return nyc_qualifying_cdcc_amount * applicable_percentage



from policyengine_us.model_api import *


class dc_ccsp_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for DC Child Care Subsidy Program (CCSP)"
    definition_period = MONTH
    defined_for = StateCode.DC
    reference = "https://osse.dc.gov/sites/default/files/dc/sites/osse/publication/attachments/DC%20Child%20Care%20Subsidy%20Program%20Policy%20Manual.pdf#page=8"

    def formula(spm_unit, period, parameters):
        person = spm_unit.members
        eligible_child = person("dc_ccsp_eligible_child", period)
        has_eligible_child = spm_unit.any(eligible_child)
        asset_eligible = spm_unit("dc_ccsp_asset_eligible", period)
        income_eligible = spm_unit("dc_ccsp_income_eligible", period)
        income_test_waived = spm_unit("dc_ccsp_income_test_waived", period)
        qualified_activity_or_need_eligible = spm_unit(
            "dc_ccsp_qualified_activity_or_need_eligible", period
        )
        return (
            has_eligible_child
            & asset_eligible
            & (income_eligible | income_test_waived)
            & qualified_activity_or_need_eligible
        )



from policyengine_us.model_api import *


class is_ssi_eligible_spouse(Variable):
    value_type = bool
    entity = Person
    label = "Is an SSI-eligible spouse"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/42/1382c#b"

    def formula(person, period, parameters):
        both_aged_blind_disabled = person("ssi_marital_both_eligible", period)
        spouse = person("is_tax_unit_spouse", period)
        return spouse & both_aged_blind_disabled



from policyengine_us.model_api import *


class az_stillborn_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "Arizona stillborn exemption"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.AZ

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.az.tax.income.exemptions

        stillborn = tax_unit("tax_unit_stillborn_children", period)

        return stillborn * p.stillborn



from policyengine_us.model_api import *


class nh_old_age_exemption(Variable):
    value_type = float
    entity = TaxUnit
    label = "New Hampshire old age exemption"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NH

    def formula(tax_unit, period, parameters):
        # Then get the NH old age exemptions part of the parameter tree.
        p = parameters(period).gov.states.nh.tax.income.exemptions

        # Check if the individual's eligibility.
        head_eligible = (
            tax_unit("age_head", period) >= p.old_age_eligibility
        ).astype(int)

        # Check if the individual spouse's eligibility.
        spouse_eligible = (
            tax_unit("age_spouse", period) >= p.old_age_eligibility
        ).astype(int)

        # Calculate total blind exemption.
        return (head_eligible + spouse_eligible) * p.amount.old_age_addition



from policyengine_us.model_api import *


class ma_taxpayer_refund_rebate(Variable):
    value_type = float
    entity = TaxUnit
    label = "Massachusetts taxpayer refund rebate"
    defined_for = StateCode.MA
    unit = USD
    definition_period = YEAR
    reference = "https://tax.hawaii.gov/act-115-ref/"

    def formula(tax_unit, period, parameters):
        income_tax_before_credits = tax_unit(
            "ma_income_tax_before_refundable_credits", period
        )
        p = parameters(period).gov.states.ma.tax.income.credits

        # Sum refundable credits except ma_taxpayer_refund_rebate to avoid circular reference
        refundable_credit_list = p.refundable
        other_refundable_credits = 0
        for credit in refundable_credit_list:
            if credit != "ma_taxpayer_refund_rebate":
                other_refundable_credits += tax_unit(credit, period)

        income_tax_less_credits = max_(
            income_tax_before_credits - other_refundable_credits, 0
        )
        return p.taxpayer_refund_rebate.rate * income_tax_less_credits



from policyengine_us.model_api import *


class in_out_of_home_care_facility(Variable):
    value_type = bool
    entity = Person
    label = "Is in a nonmedical out of home care facility"
    definition_period = YEAR



from policyengine_us.model_api import *


class dwks09(Variable):
    value_type = float
    entity = TaxUnit
    definition_period = YEAR
    label = "IRS Form 1040 Schedule D worksheet (part 2 of 6)"
    unit = USD

    def formula(tax_unit, period, parameters):
        # SchD lines 15 and 16, respectively
        qdiv_plus_ltcg = add(
            tax_unit,
            period,
            ["long_term_capital_gains", "qualified_dividend_income"],
        )
        net_cg = min_(qdiv_plus_ltcg, tax_unit("net_capital_gains", period))
        # dwks08 = min(dwks03, dwks04)
        # dwks09 = max(0., dwks07 - dwks08)
        # BELOW TWO STATEMENTS ARE UNCLEAR IN LIGHT OF dwks09=... COMMENT
        other_cg = add(tax_unit, period, ["non_sch_d_capital_gains"])
        mod_cg = where(other_cg > 0, other_cg, max_(0, net_cg) + other_cg)
        return max_(
            0,
            mod_cg - min_(0, tax_unit("investment_income_form_4952", period)),
        )



from policyengine_us.model_api import *


class ri_subtractions(Variable):
    value_type = float
    entity = TaxUnit
    label = "Rhode Island AGI Subtractions"
    unit = USD
    definition_period = YEAR
    reference = "https://tax.ri.gov/sites/g/files/xkgbur541/files/2022-12/2022%201041%20Schedule%20M_w.pdf#page=1"
    defined_for = StateCode.RI

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ri.tax.income.agi.subtractions
        total_subtractions = add(tax_unit, period, p.subtractions)
        # Prevent negative subtractions from acting as additions
        return max_(0, total_subtractions)



from policyengine_us.model_api import *


class charitable_cash_donations(Variable):
    value_type = float
    entity = Person
    label = "Charitable donations (cash)"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class il_liheap_eligible(Variable):
    value_type = bool
    entity = SPMUnit
    label = "Eligible for Illinois LIHEAP"
    definition_period = YEAR
    defined_for = "il_liheap_income_eligible"
    reference = "https://dceo.illinois.gov/communityservices/utilitybillassistance.html"

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.il.dceo.liheap.eligibility

        # For renters with heat included in rent, check additional requirement
        heat_in_rent = spm_unit("heat_expense_included_in_rent", period)
        rent = add(spm_unit, period, ["rent"])
        income = add(spm_unit, period, ["irs_gross_income"])

        # If heat is included in rent, rent must exceed threshold percentage of income
        rent_threshold = income * p.rent_rate
        rent_threshold_met = rent > rent_threshold

        return ~heat_in_rent | rent_threshold_met



from policyengine_us.model_api import *


class federal_state_income_tax(Variable):
    value_type = float
    entity = TaxUnit
    label = "Total federal and state income tax"
    unit = USD
    definition_period = YEAR

    adds = ["income_tax", "state_income_tax"]



from policyengine_us.model_api import *


class ne_cdcc_nonrefundable(Variable):
    value_type = float
    entity = TaxUnit
    label = "Nebraska nonrefundable cdcc"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://revenue.nebraska.gov/files/doc/tax-forms/2021/f_1040n_booklet.pdf"
        "https://revenue.nebraska.gov/files/doc/2022_Ne_Individual_Income_Tax_Booklet_8-307-2022_final_5.pdf"
    )
    defined_for = StateCode.NE

    def formula(tax_unit, period, parameters):
        p = parameters(period).gov.states.ne.tax.income.credits
        # determine AGI eligibility
        eligible = ~tax_unit("ne_cdcc_refundable_eligible", period)
        # determine NE nonrefundable cdcc amount
        # Nebraska matches the potential federal credit
        us_cdcc = tax_unit("cdcc_potential", period)
        ne_cdcc = us_cdcc * p.cdcc.nonrefundable.fraction
        return eligible * ne_cdcc



from policyengine_us.model_api import *


class me_eitc(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maine EITC"
    unit = USD
    definition_period = YEAR
    reference = "https://www.mainelegislature.org/legis/statutes/36/title36sec5219-S.html"
    defined_for = StateCode.ME

    def formula(tax_unit, period, parameters):
        federal_eitc = tax_unit("eitc", period)
        p = parameters(period).gov.states.me.tax.income.credits.eitc
        match_rate = where(
            tax_unit("eitc_child_count", period) > 0,
            p.rate.with_qualifying_child,
            p.rate.no_qualifying_child,
        )
        return federal_eitc * match_rate



from policyengine_us.model_api import *


class il_hbwd_countable_assets(Variable):
    value_type = float
    unit = USD
    entity = Person
    label = "Illinois Health Benefits for Workers with Disabilities countable assets"
    definition_period = MONTH
    reference = (
        "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-120.381",
        "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-120.510",
        "https://hfs.illinois.gov/medicalprograms/hbwd/eligibility.html",
    )
    defined_for = StateCode.IL

    def formula(person, period, parameters):
        # HBWD exempt assets per § 120.381:
        # - Primary residence (homestead)
        # - Personal effects and household goods
        # - One vehicle (full exemption if essential, $4,500 if not)
        # - Life insurance ≤$1,500
        # - Burial resources
        # - Retirement accounts inaccessible before age 59.5
        # - Medical savings accounts (26 USC 220)
        # Currently only implementing vehicle exemption
        # § 120.381 and § 113.141 have identical vehicle exemption rules,
        # so using AABD vehicle calculation
        cash_assets = person.spm_unit("spm_unit_cash_assets", period.this_year)
        vehicle_value = person.spm_unit(
            "il_aabd_countable_vehicle_value", period
        )
        # TODO: Implement homestead, retirement account, burial, and other exemptions
        return cash_assets + vehicle_value



from policyengine_us.model_api import *


class co_pension_subtraction(Variable):
    value_type = float
    entity = TaxUnit
    label = "Colorado pension and annuity subtraction"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.CO

    adds = [
        "co_pension_subtraction_indv",
        "co_social_security_subtraction_indv",
    ]



from policyengine_us.model_api import *


class heating_expense_person(Variable):
    value_type = float
    entity = Person
    label = "Heating cost for each person"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class salt_refund_income(Variable):
    value_type = float
    entity = Person
    label = "State and local tax refund income"
    unit = USD
    definition_period = YEAR



from policyengine_us.model_api import *


class marital_unit_weight(Variable):
    value_type = float
    entity = MaritalUnit
    label = "Marital unit weight"
    definition_period = YEAR

    def formula(marital_unit, period, parameters):
        # If no weight provided, use the average of the people in the marital unit.
        return marital_unit.household("household_weight", period)



from policyengine_us.model_api import *


class StateFilingStatusIfMarriedFilingSeparatelyOnSameReturn(Enum):
    SINGLE = "Single"
    SEPARATE = "Separate"
    HEAD_OF_HOUSEHOLD = "Head of household"
    SURVIVING_SPOUSE = "Surviving spouse"


class state_filing_status_if_married_filing_separately_on_same_return(
    Variable
):
    value_type = Enum
    entity = TaxUnit
    possible_values = StateFilingStatusIfMarriedFilingSeparatelyOnSameReturn
    default_value = (
        StateFilingStatusIfMarriedFilingSeparatelyOnSameReturn.SINGLE
    )
    definition_period = YEAR
    label = "State filing status for the tax unit if married couple file separately on same return"

    def formula(tax_unit, period, parameters):
        us_filing_status = tax_unit("filing_status", period)
        fsvals = us_filing_status.possible_values
        return select(
            [
                us_filing_status == fsvals.JOINT,
                us_filing_status == fsvals.SINGLE,
                us_filing_status == fsvals.SEPARATE,
                us_filing_status == fsvals.HEAD_OF_HOUSEHOLD,
                us_filing_status == fsvals.SURVIVING_SPOUSE,
            ],
            [
                # Simulate scenario where joint filers file separately on Arkansas return.
                StateFilingStatusIfMarriedFilingSeparatelyOnSameReturn.SEPARATE,
                StateFilingStatusIfMarriedFilingSeparatelyOnSameReturn.SINGLE,
                StateFilingStatusIfMarriedFilingSeparatelyOnSameReturn.SEPARATE,
                StateFilingStatusIfMarriedFilingSeparatelyOnSameReturn.HEAD_OF_HOUSEHOLD,
                StateFilingStatusIfMarriedFilingSeparatelyOnSameReturn.SURVIVING_SPOUSE,
            ],
        )



from policyengine_us.model_api import *


class ca_tanf_db_unearned_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "California CalWORKs Disability-Based Unearned Income"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.CA
    reference = "http://epolicy.dpss.lacounty.gov/epolicy/epolicy/server/general/projects_responsive/ePolicyMaster/index.htm?&area=general&type=responsivehelp&ctxid=&project=ePolicyMaster#t=mergedProjects%2FCalWORKs%2FCalWORKs%2F44-101_Income_Definitions%2F44-101_Income_Definitions.htm%23Definitionsbc-4&rhtocid=_3_1_6_0_3"

    adds = "gov.states.ca.cdss.tanf.cash.income.sources.db_unearned"



from policyengine_us.model_api import *


class ny_cdcc_rate(Variable):
    value_type = float
    entity = TaxUnit
    label = "NY CDCC rate"
    unit = USD
    definition_period = YEAR
    reference = "https://www.nysenate.gov/legislation/laws/TAX/606"  # (c)
    defined_for = StateCode.NY

    def formula(tax_unit, period, parameters):
        # Under the law (s. 606(1)), the NY CDCC defined at its core as:
        # (federal CDCC) x (some percentage)
        # There is also another important part: s. (1-b), which amends this to:
        # (federal CDCC) x (some percentage) x (some other percentage based on NY AGI)
        # The last complication is that in (1) people with income under $40,000 are
        # treated differently than people with income over $40,000.

        # First, we'll calculate the first percentage. We calculate the percentage
        # for people under $40k, and then the percentage if over $40k, and then use the
        # correct one with a final check.

        # The core logic for both income groups is the same:
        # percentage = "A%" + (greater of: $B or ($C - NY AGI)) / $D
        # where A, B, C and D are parameters set by law. The law doesn't give
        # any intuitive names for these parameters.

        ny_agi = tax_unit("ny_agi", period)
        percentage = parameters(
            period
        ).gov.states.ny.tax.income.credits.cdcc.percentage
        main = percentage.main
        main_numerator = main.fraction.numerator
        fraction = (
            max_(0, min_(main_numerator.min, main_numerator.top - ny_agi))
            / main.fraction.denominator
        )
        main_percentage = main.base_percentage + fraction * main.multiplier

        alternate = percentage.alternate
        alternate_numerator = alternate.fraction.numerator
        fraction = (
            max_(
                0,
                min_(
                    alternate_numerator.min, alternate_numerator.top - ny_agi
                ),
            )
            / alternate.fraction.denominator
        )
        alternate_percentage = (
            alternate.base_percentage + fraction * alternate.multiplier
        )

        ny_percentage = where(
            ny_agi < alternate.max_agi,
            alternate_percentage,
            main_percentage,
        )

        s_1_b_multiplier = percentage.multiplier.calc(ny_agi)

        return ny_percentage * s_1_b_multiplier



from policyengine_us.model_api import *
from policyengine_core.periods import period as period_
from policyengine_core.periods import instant
import numpy as np


def create_afa_other_dependent_credit() -> Reform:
    class other_dependent_credit(Variable):
        value_type = float
        entity = TaxUnit
        label = "Other Dependent Credit which is independent of the CTC"
        unit = USD
        documentation = (
            "The standalone dependent credit in respect of this person."
        )
        definition_period = YEAR
        reference = "https://www.bennet.senate.gov/wp-content/uploads/2025/04/American-Family-Act-2025.pdf"

        def formula(tax_unit, period, parameters):
            maximum_amount = add(
                tax_unit, period, ["other_dependent_credit_maximum"]
            )
            reduction = tax_unit("other_dependent_credit_phase_out", period)
            return max_(0, maximum_amount - reduction)

    class other_dependent_credit_maximum(Variable):
        value_type = float
        entity = Person
        label = "Other Dependent Credit Maximum"
        unit = USD
        definition_period = YEAR
        reference = "https://www.bennet.senate.gov/wp-content/uploads/2025/04/American-Family-Act-2025.pdf"

        def formula(person, period, parameters):
            p = parameters(
                period
            ).gov.contrib.congress.afa.other_dependent_credit
            is_dependent = person("is_tax_unit_dependent", period)
            is_child = person("ctc_child_individual_maximum", period) > 0
            is_adult_dependent = ~is_child & is_dependent
            return is_adult_dependent * p.amount

    class other_dependent_credit_phase_out(Variable):
        value_type = float
        entity = TaxUnit
        label = "Other Dependent Credit Phase-Out"
        unit = USD
        definition_period = YEAR
        reference = "https://www.bennet.senate.gov/wp-content/uploads/2025/04/American-Family-Act-2025.pdf"

        def formula(tax_unit, period, parameters):
            # TCJA's phase-out changes are purely parametric so don't require
            # structural reform.

            # The ARPA CTC has two phase-outs: the original, and a new phase-out
            # applying before and only to the increase in the maximum CTC under ARPA.

            # Start with the normal phase-out.
            income = tax_unit("adjusted_gross_income", period)
            p = parameters(period).gov.contrib.congress.afa.ctc.phase_out
            filing_status = tax_unit("filing_status", period)
            phase_out_threshold = p.threshold.higher[filing_status]
            excess = max_(0, income - phase_out_threshold)
            increments = np.ceil(excess / p.increment)
            return increments * p.amount

    class ctc_lower_phase_out(Variable):
        value_type = float
        entity = TaxUnit
        label = "CTC reduction from income"
        unit = USD
        documentation = "Reduction of the total CTC due to income."
        definition_period = YEAR
        reference = "https://www.bennet.senate.gov/wp-content/uploads/2025/04/American-Family-Act-2025.pdf"

        def formula(tax_unit, period, parameters):
            # TCJA's phase-out changes are purely parametric so don't require
            # structural reform.

            # The ARPA CTC has two phase-outs: the original, and a new phase-out
            # applying before and only to the increase in the maximum CTC under ARPA.

            # Start with the normal phase-out.
            income = tax_unit("adjusted_gross_income", period)
            p = parameters(period).gov.contrib.congress.afa.ctc.phase_out
            filing_status = tax_unit("filing_status", period)
            lower_threshold = p.threshold.lower[filing_status]
            excess = max_(0, income - lower_threshold)
            increments = np.ceil(excess / p.increment)
            return increments * p.amount

    class ctc_higher_phase_out(Variable):
        value_type = float
        entity = TaxUnit
        label = "CTC reduction from income"
        unit = USD
        documentation = "Reduction of the total CTC due to income."
        definition_period = YEAR
        reference = "https://www.bennet.senate.gov/wp-content/uploads/2025/04/American-Family-Act-2025.pdf"

        def formula(tax_unit, period, parameters):
            # TCJA's phase-out changes are purely parametric so don't require
            # structural reform.

            # The ARPA CTC has two phase-outs: the original, and a new phase-out
            # applying before and only to the increase in the maximum CTC under ARPA.

            # Start with the normal phase-out.
            income = tax_unit("adjusted_gross_income", period)
            p = parameters(period).gov.contrib.congress.afa.ctc.phase_out
            filing_status = tax_unit("filing_status", period)
            lower_threshold = p.threshold.higher[filing_status]
            excess = max_(0, income - lower_threshold)
            increments = np.ceil(excess / p.increment)
            return increments * p.amount

    class refundable_ctc(Variable):
        value_type = float
        entity = TaxUnit
        label = "refundable CTC"
        unit = USD
        documentation = (
            "The portion of the Child Tax Credit that is refundable."
        )
        definition_period = YEAR
        reference = "https://www.bennet.senate.gov/wp-content/uploads/2025/04/American-Family-Act-2025.pdf"

        def formula(tax_unit, period, parameters):
            # This line corresponds to "the credit which would be allowed under this section [the CTC section]"
            # without regard to this subsection [the refundability section] and the limitation under
            # section 26(a) [the section that limits the amount of the non-refundable CTC to tax liability].
            # This is the full CTC. This is then limited to the maximum refundable amount per child as per the
            # TCJA provision.

            ctc = parameters(period).gov.irs.credits.ctc

            maximum_amount = tax_unit("ctc_refundable_maximum", period)
            total_ctc = tax_unit("ctc", period)

            if ctc.refundable.fully_refundable:
                reduction = tax_unit("ctc_lower_phase_out", period)
                p = parameters(period).gov.contrib.congress.afa.ctc.phase_out
                qualifying_children = tax_unit(
                    "ctc_qualifying_children", period
                )
                lower_floor = p.lower_floor * qualifying_children
                capped_credit = min_(lower_floor, maximum_amount)
                reduced_max_amount_lower = max_(
                    capped_credit, maximum_amount - reduction
                )
                higher_reduction = tax_unit("ctc_higher_phase_out", period)
                reduced_max_amount = max_(
                    0, reduced_max_amount_lower - higher_reduction
                )
                return min_(reduced_max_amount, total_ctc)

            maximum_refundable_ctc = min_(maximum_amount, total_ctc)

            phase_in = tax_unit("ctc_phase_in", period)
            limiting_tax = tax_unit("ctc_limiting_tax_liability", period)
            ctc_capped_by_tax = min_(total_ctc, limiting_tax)
            ctc_capped_by_increased_tax = min_(
                total_ctc, limiting_tax + phase_in
            )
            amount_ctc_would_increase = (
                ctc_capped_by_increased_tax - ctc_capped_by_tax
            )
            return min_(maximum_refundable_ctc, amount_ctc_would_increase)

    class ctc(Variable):
        value_type = float
        entity = TaxUnit
        label = "Child Tax Credit"
        unit = USD
        documentation = "Total value of the non-refundable and refundable portions of the Child Tax Credit."
        definition_period = YEAR
        reference = "https://www.bennet.senate.gov/wp-content/uploads/2025/04/American-Family-Act-2025.pdf"

        def formula(tax_unit, period, parameters):
            maximum_amount = add(
                tax_unit, period, ["ctc_child_individual_maximum_arpa"]
            )
            reduction = tax_unit("ctc_lower_phase_out", period)
            p = parameters(period).gov.contrib.congress.afa.ctc.phase_out
            qualifying_children = tax_unit("ctc_qualifying_children", period)
            lower_floor = p.lower_floor * qualifying_children
            reduced_max_amount_lower = max_(
                lower_floor, maximum_amount - reduction
            )
            higher_reduction = tax_unit("ctc_higher_phase_out", period)
            return max_(0, reduced_max_amount_lower - higher_reduction)

    class ctc_child_individual_maximum_arpa(Variable):
        value_type = float
        entity = Person
        label = "CTC maximum amount (child under ARPA)"
        unit = USD
        documentation = "The CTC entitlement in respect of this person as a child, under the American Rescue Plan Act."
        definition_period = YEAR
        reference = "https://www.bennet.senate.gov/wp-content/uploads/2025/04/American-Family-Act-2025.pdf"

        def formula(person, period, parameters):
            p = parameters(period).gov.contrib.congress.afa.ctc.amount
            base_amount = p.base
            age = person("age", period)
            multiplier = p.multiplier.calc(age)
            pre_baby_bonus_amount = base_amount * multiplier
            is_baby = age < 1
            monthly_base_amount = base_amount / MONTHS_IN_YEAR
            baby_bonus = (
                monthly_base_amount * p.baby_bonus
                - monthly_base_amount * multiplier
            )
            baby_bonus_amount = baby_bonus * is_baby
            return pre_baby_bonus_amount + baby_bonus_amount

    def modify_parameters(parameters):
        parameters.gov.irs.credits.non_refundable.update(
            start=instant("2025-01-01"),
            stop=instant("2039-12-31"),
            value=[
                "cdcc",
                "elderly_disabled_credit",
                "non_refundable_american_opportunity_credit",
                "lifetime_learning_credit",
                "savers_credit",
                "residential_clean_energy_credit",
                "energy_efficient_home_improvement_credit",
                "new_clean_vehicle_credit",
                "used_clean_vehicle_credit",
                "other_dependent_credit",
            ],
        )
        parameters.gov.irs.credits.ctc.refundable.fully_refundable.update(
            start=instant("2025-01-01"),
            stop=instant("2039-12-31"),
            value=True,
        )
        parameters.gov.irs.credits.ctc.amount.base[1].threshold.update(
            start=instant("2025-01-01"),
            stop=instant("2039-12-31"),
            value=18,
        )
        parameters.gov.irs.credits.ctc.amount.adult_dependent.update(
            start=instant("2025-01-01"),
            stop=instant("2039-12-31"),
            value=0,
        )
        return parameters

    class reform(Reform):
        def apply(self):
            self.update_variable(other_dependent_credit_maximum)
            self.update_variable(other_dependent_credit_phase_out)
            self.update_variable(other_dependent_credit)
            self.update_variable(ctc_lower_phase_out)
            self.update_variable(ctc_higher_phase_out)
            self.update_variable(ctc)
            self.update_variable(refundable_ctc)
            self.update_variable(ctc_child_individual_maximum_arpa)
            self.modify_parameters(modify_parameters)

    return reform


def create_afa_other_dependent_credit_reform(
    parameters, period, bypass: bool = False
):
    if bypass:
        return create_afa_other_dependent_credit()

    p = parameters.gov.contrib.congress.afa
    reform_active = False
    current_period = period_(period)

    for i in range(10):
        if p.in_effect(current_period):
            reform_active = True
            break
        current_period = current_period.offset(1, "year")

    if reform_active:
        return create_afa_other_dependent_credit()
    else:
        return None


afa_other_dependent_credit = create_afa_other_dependent_credit_reform(
    None, None, bypass=True
)



from policyengine_us.model_api import *


class pa_nontaxable_pension_income(Variable):
    value_type = float
    entity = Person
    label = "Pension income taxable by US but not by PA"
    unit = USD
    documentation = "US taxable pension income excluded from PA AGI."
    definition_period = YEAR
    reference = "https://www.revenue.pa.gov/FormsandPublications/FormsforIndividuals/PIT/Documents/2021/2021_pa-40in.pdf#page=8"
    defined_for = StateCode.PA

    def formula(person, period, parameters):
        retired = person("is_retired", period)
        us_taxable_pension = person("taxable_pension_income", period)
        return where(retired, us_taxable_pension, 0)



from policyengine_us.model_api import *


class us_bonds_for_higher_ed(Variable):
    value_type = float
    entity = Person
    label = "Income from U.S. bonds spent on higher education"
    unit = USD
    definition_period = YEAR
    reference = "https://www.law.cornell.edu/uscode/text/26/135"



from policyengine_us.model_api import *


class ia_alternate_tax_indiv(Variable):
    value_type = float
    entity = Person
    label = "Iowa alternate tax when married couples file separately"
    unit = USD
    definition_period = YEAR
    reference = (
        "https://tax.iowa.gov/sites/default/files/2022-01/IA1040%2841-001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-01/2021%20Expanded%20Instructions_010323.pdf#page=53"
        "https://tax.iowa.gov/sites/default/files/2023-01/2022IA1040%2841001%29.pdf"
        "https://tax.iowa.gov/sites/default/files/2023-03/2022%20Expanded%20Instructions_022023.pdf#page=53"
    )
    defined_for = StateCode.IA

    def formula(person, period, parameters):
        alt_tax = person.tax_unit("ia_alternate_tax_unit", period)
        return alt_tax * person("ia_prorate_fraction", period)



from policyengine_us.model_api import *


class ny_cdcc_max(Variable):
    value_type = float
    entity = TaxUnit
    label = "Maximum NY CDCC"
    unit = USD
    definition_period = YEAR
    defined_for = StateCode.NY
    reference = "https://www.nysenate.gov/legislation/laws/TAX/606"  # (c)

    def formula(tax_unit, period, parameters):
        ny_cdcc = parameters(period).gov.states.ny.tax.income.credits.cdcc
        count_eligible = tax_unit("count_cdcc_eligible", period)
        cdcc_expenses = tax_unit("cdcc_relevant_expenses", period)
        lower_earnings = tax_unit("min_head_spouse_earned", period)
        ny_cap = ny_cdcc.max.calc(count_eligible)
        return min_(cdcc_expenses, min_(ny_cap, lower_earnings))



from policyengine_us.model_api import *


class mi_homestead_property_tax_credit_eligible(Variable):
    value_type = bool
    entity = TaxUnit
    label = "Eligible for the Michigan homestead property tax credit"
    definition_period = YEAR
    reference = (
        "http://legislature.mi.gov/doc.aspx?mcl-206-508",
        "https://www.michigan.gov/taxes/-/media/Project/Websites/taxes/Forms/2022/2022-IIT-Forms/MI-1040CR.pdf#page=1",
        "https://www.legislature.mi.gov/Publications/TaxpayerGuide.pdf#page=16",
    )
    defined_for = StateCode.MI

    def formula(tax_unit, period, parameters):
        p = parameters(
            period
        ).gov.states.mi.tax.income.credits.homestead_property_tax
        # Line 38 & 39 & 41
        return (
            add(tax_unit, period, ["assessed_property_value"])
            <= p.property_value_limit
        )



from policyengine_us.model_api import *


class dc_tanf_countable_unearned_income(Variable):
    value_type = float
    entity = SPMUnit
    label = "DC Temporary Assistance for Needy Families (TANF) countable unearned income"
    unit = USD
    definition_period = MONTH
    reference = "https://code.dccouncil.gov/us/dc/council/code/sections/4-205.11"  # (a)(8)
    defined_for = StateCode.DC

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.dc.dhs.tanf.income
        total_unearned_income = add(
            spm_unit, period, ["dc_tanf_gross_unearned_income"]
        )
        child_support = add(spm_unit, period, ["child_support_received"])
        child_support_deduction = min_(
            child_support, p.deductions.child_support
        )

        return max_(0, total_unearned_income - child_support_deduction)



from policyengine_us.model_api import *


class employer_medicare_tax(Variable):
    value_type = float
    entity = Person
    label = "Employer-side health insurance payroll tax"
    documentation = (
        "Total liability for employer-side health insurance payroll tax."
    )
    definition_period = YEAR
    unit = USD

    def formula(person, period, parameters):
        p = parameters(period).gov.irs.payroll.medicare.rate
        return p.employer * person("payroll_tax_gross_wages", period)



from policyengine_us.model_api import *


class il_tanf_payment_level_for_grant_calculation(Variable):
    value_type = float
    entity = SPMUnit
    label = "Illinois Temporary Assistance for Needy Families (TANF) payment level for grant calculation"
    unit = USD
    definition_period = MONTH
    reference = "https://www.law.cornell.edu/regulations/illinois/Ill-Admin-Code-tit-89-SS-112.251"
    defined_for = StateCode.IL

    def formula(spm_unit, period, parameters):
        p = parameters(period).gov.states.il.dhs.tanf.payment_level
        fpg = spm_unit("il_tanf_assistance_unit_fpg", period)
        parent_count = add(
            spm_unit, period, ["il_tanf_payment_eligible_parent"]
        )
        child_count = add(spm_unit, period, ["il_tanf_payment_eligible_child"])

        parent_only = (parent_count > 0) & (child_count == 0)
        child_only = (child_count > 0) & (parent_count == 0)
        parent_and_child_present = (parent_count > 0) & (child_count > 0)
        # The payment levels can vary by ssi eligibility:
        # If only the parent is receiving SSI, the payment level is 75%
        # if only the child is receiving SSI, the payment level is 25%
        base_payment = p.rate * fpg
        parent_only_payment = p.parent_only_rate * base_payment
        child_only_payment = p.child_only_rate * base_payment

        return select(
            [parent_only, child_only, parent_and_child_present],
            [
                parent_only_payment,
                child_only_payment,
                base_payment,
            ],
            default=0,
        )
